from datetime import date, timedelta
from decimal import Decimal

from tests.testcase import BaseTestCase


class LowCardinalityTestCase(BaseTestCase):
    required_server_version = (19, 3, 3)
    stable_support_version = (19, 9, 2)

    def cli_client_kwargs(self):
        if self.server_version >= self.stable_support_version:
            return {'allow_suspicious_low_cardinality_types': 1}

    def test_uint8(self):
        with self.create_table('a LowCardinality(UInt8)'):
            data = [(x, ) for x in range(255)]
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                '\n'.join(str(x[0]) for x in data) + '\n'
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_int8(self):
        with self.create_table('a LowCardinality(Int8)'):
            data = [(x - 127, ) for x in range(255)]
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                '\n'.join(str(x[0]) for x in data) + '\n'

            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_nullable_int8(self):
        with self.create_table('a LowCardinality(Nullable(Int8))'):
            data = [(None, ), (-1, ), (0, ), (1, ), (None, )]
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '\\N\n-1\n0\n1\n\\N\n')

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_date(self):
        with self.create_table('a LowCardinality(Date)'):
            start = date(1970, 1, 1)
            data = [(start + timedelta(x), ) for x in range(300)]
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_float(self):
        with self.create_table('a LowCardinality(Float)'):
            data = [(float(x),) for x in range(300)]
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_decimal(self):
        with self.create_table('a LowCardinality(Float)'):
            data = [(Decimal(x),) for x in range(300)]
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_array(self):
        with self.create_table('a Array(LowCardinality(Int16))'):
            data = [([100, 500], )]
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '[100,500]\n')

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_empty_array(self):
        with self.create_table('a Array(LowCardinality(Int16))'):
            data = [([], )]
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '[]\n')

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_string(self):
        with self.create_table('a LowCardinality(String)'):
            data = [
                ('test', ), ('low', ), ('cardinality', ),
                ('test', ), ('test', ), ('', )
            ]
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                'test\nlow\ncardinality\ntest\ntest\n\n'
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_fixed_string(self):
        with self.create_table('a LowCardinality(FixedString(12))'):
            data = [
                ('test', ), ('low', ), ('cardinality', ),
                ('test', ), ('test', ), ('', )
            ]
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                'test\\0\\0\\0\\0\\0\\0\\0\\0\n'
                'low\\0\\0\\0\\0\\0\\0\\0\\0\\0\n'
                'cardinality\\0\n'
                'test\\0\\0\\0\\0\\0\\0\\0\\0\n'
                'test\\0\\0\\0\\0\\0\\0\\0\\0\n'
                '\\0\\0\\0\\0\\0\\0\\0\\0\\0\\0\\0\\0\n'
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_nullable_string(self):
        with self.create_table('a LowCardinality(Nullable(String))'):
            data = [
                ('test', ), ('', ), (None, )
            ]
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                'test\n\n\\N\n'
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)



from tests.testcase import BaseTestCase
from clickhouse_driver import errors
from tests.util import require_server_version


class IntTestCase(BaseTestCase):
    def test_chop_to_type(self):
        with self.create_table('a UInt8'):
            data = [(300, )]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, types_check=True
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '44\n')
            inserted = self.client.execute(query)
            self.assertEqual(inserted, [(44, )])

        with self.create_table('a Int8'):
            data = [(-300,)]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, types_check=True
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '-44\n')
            inserted = self.client.execute(query)
            self.assertEqual(inserted, [(-44, )])

    def test_raise_struct_error(self):
        with self.create_table('a UInt8'):
            with self.assertRaises(errors.TypeMismatchError) as e:
                data = [(300, )]
                self.client.execute(
                    'INSERT INTO test (a) VALUES', data
                )

            exc = str(e.exception)
            self.assertIn('Column a', exc)
            self.assertIn('types_check=True', exc)

    def test_uint_type_mismatch(self):
        data = [(-1, )]
        with self.create_table('a UInt8'):
            with self.assertRaises(errors.TypeMismatchError) as e:
                self.client.execute(
                    'INSERT INTO test (a) VALUES', data, types_check=True
                )

            self.assertIn('-1 for column "a"', str(e.exception))

            with self.assertRaises(errors.TypeMismatchError) as e:
                self.client.execute(
                    'INSERT INTO test (a) VALUES', data
                )

            self.assertIn('Column a', str(e.exception))

    def test_all_sizes(self):
        columns = (
            'a Int8, b Int16, c Int32, d Int64, '
            'e UInt8, f UInt16, g UInt32, h UInt64'
        )

        data = [
            (-10, -300, -123581321, -123581321345589144,
             10, 300, 123581321, 123581321345589144)
        ]
        with self.create_table(columns):
            self.client.execute(
                'INSERT INTO test (a, b, c, d, e, f, g, h) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted, (
                    '-10\t-300\t-123581321\t-123581321345589144\t'
                    '10\t300\t123581321\t123581321345589144\n'
                )
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_corner_cases(self):
        columns = (
            'a Int8, b Int16, c Int32, d Int64, '
            'e UInt8, f UInt16, g UInt32, h UInt64'
        )

        data = [
            (-128, -32768, -2147483648, -9223372036854775808,
             255, 65535, 4294967295, 18446744073709551615),
            (127, 32767, 2147483647, 9223372036854775807, 0, 0, 0, 0),
        ]
        with self.create_table(columns):
            self.client.execute(
                'INSERT INTO test (a, b, c, d, e, f, g, h) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted, (
                    '-128\t-32768\t-2147483648\t-9223372036854775808\t'
                    '255\t65535\t4294967295\t18446744073709551615\n'
                    '127\t32767\t2147483647\t9223372036854775807\t0\t0\t0\t0\n'
                )
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_nullable(self):
        with self.create_table('a Nullable(Int32)'):
            data = [(2, ), (None, ), (4, ), (None, ), (8, )]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '2\n\\N\n4\n\\N\n8\n')

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)


class BigIntTestCase(BaseTestCase):
    required_server_version = (20, 8, 2)

    def cli_client_kwargs(self):
        return {'allow_experimental_bigint_types': 1}

    def test_int128(self):
        with self.create_table('a Int128'):
            data = [
                (-170141183460469231731687303715884105728, ),
                (-111111111111111111111111111111111111111, ),
                (123, ),
                (111111111111111111111111111111111111111, ),
                (170141183460469231731687303715884105727, )
            ]
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                '-170141183460469231731687303715884105728\n'
                '-111111111111111111111111111111111111111\n'
                '123\n'
                '111111111111111111111111111111111111111\n'
                '170141183460469231731687303715884105727\n'
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    @require_server_version(21, 6)
    def test_uint128(self):
        with self.create_table('a UInt128'):
            data = [
                (0, ),
                (123, ),
                (340282366920938463463374607431768211455, )
            ]
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                '0\n'
                '123\n'
                '340282366920938463463374607431768211455\n'
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_int256(self):
        with self.create_table('a Int256'):
            data = [
                (-57896044618658097711785492504343953926634992332820282019728792003956564819968, ),  # noqa: E501
                (-11111111111111111111111111111111111111111111111111111111111111111111111111111, ),  # noqa: E501
                (123, ),
                (11111111111111111111111111111111111111111111111111111111111111111111111111111, ),  # noqa: E501
                (57896044618658097711785492504343953926634992332820282019728792003956564819967, )  # noqa: E501
            ]
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                '-57896044618658097711785492504343953926634992332820282019728792003956564819968\n'  # noqa: E501
                '-11111111111111111111111111111111111111111111111111111111111111111111111111111\n'  # noqa: E501
                '123\n'
                '11111111111111111111111111111111111111111111111111111111111111111111111111111\n'  # noqa: E501
                '57896044618658097711785492504343953926634992332820282019728792003956564819967\n'  # noqa: E501
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_uint256(self):
        with self.create_table('a UInt256'):
            data = [
                (0, ),
                (123, ),
                (111111111111111111111111111111111111111111111111111111111111111111111111111111, ),  # noqa: E501
                (115792089237316195423570985008687907853269984665640564039457584007913129639935, )  # noqa: E501
            ]
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                '0\n'
                '123\n'
                '111111111111111111111111111111111111111111111111111111111111111111111111111111\n'  # noqa: E501
                '115792089237316195423570985008687907853269984665640564039457584007913129639935\n'  # noqa: E501
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)



from enum import IntEnum

from tests.testcase import BaseTestCase


class SimpleAggregateFunctionTestCase(BaseTestCase):
    required_server_version = (19, 8, 3)

    def test_simple(self):
        columns = 'a SimpleAggregateFunction(any, Int32)'

        data = [(3, ), (2, )]
        with self.create_table(columns):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted, '3\n2\n'
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_nullable(self):
        columns = 'a SimpleAggregateFunction(any, Nullable(Int32))'

        data = [(3, ), (None, ), (2, )]
        with self.create_table(columns):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted, '3\n\\N\n2\n'
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_simple_agg_function(self):
        class A(IntEnum):
            hello = -1
            world = 2

        columns = "a SimpleAggregateFunction(anyLast, " \
                  "Enum8('hello' = -1, 'world' = 2))"

        data = [(A.hello,), (A.world,), (-1,), (2,)]
        with self.create_table(columns):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted, (
                    'hello\n'
                    'world\n'
                    'hello\n'
                    'world\n'
                )
            )

            inserted = self.client.execute(query)
            self.assertEqual(
                inserted, [
                    ('hello',), ('world',),
                    ('hello',), ('world',)
                ]
            )

    def test_simple_agg_function_nullable(self):
        class A(IntEnum):
            hello = -1
            world = 2

        columns = "a SimpleAggregateFunction(anyLast, " \
                  "Nullable(Enum8('hello' = -1, 'world' = 2)))"

        data = [(A.hello,), (A.world,), (None,), (-1,), (2,)]
        with self.create_table(columns):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted, (
                    'hello\n'
                    'world\n'
                    '\\N\n'
                    'hello\n'
                    'world\n'
                )
            )

            inserted = self.client.execute(query)
            self.assertEqual(
                inserted, [
                    ('hello',), ('world',),
                    (None, ),
                    ('hello',), ('world',)
                ]
            )



from tests.testcase import BaseTestCase
from tests.util import require_server_version
from clickhouse_driver.columns import nestedcolumn


class NestedTestCase(BaseTestCase):
    def entuple(self, lst):
        return tuple(
            self.entuple(x) if isinstance(x, list) else x for x in lst
        )

    @require_server_version(21, 3, 13)
    def test_simple(self):
        columns = 'n Nested(i Int32, s String)'

        # INSERT INTO test_nested VALUES ([(0, 'a'), (1, 'b')]);
        data = [([(0, 'a'), (1, 'b')],)]

        with self.create_table(columns, flatten_nested=0):
            self.client.execute(
                'INSERT INTO test (n) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, "[(0,'a'),(1,'b')]\n")

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

            projected_i = self.client.execute('SELECT n.i FROM test')
            self.assertEqual(
                projected_i,
                [([0, 1],)]
            )

            projected_s = self.client.execute('SELECT n.s FROM test')
            self.assertEqual(
                projected_s,
                [(['a', 'b'],)]
            )

    @require_server_version(21, 3, 13)
    def test_multiple_rows(self):
        columns = 'n Nested(i Int32, s String)'

        data = [([(0, 'a'), (1, 'b')],), ([(3, 'd'), (4, 'e')],)]

        with self.create_table(columns, flatten_nested=0):
            self.client.execute(
                'INSERT INTO test (n) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                "[(0,'a'),(1,'b')]\n[(3,'d'),(4,'e')]\n"
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    @require_server_version(21, 3, 13)
    def test_dict(self):
        columns = 'n Nested(i Int32, s String)'

        data = [
            {'n': [{'i': 0, 's': 'a'}, {'i': 1, 's': 'b'}]},
            {'n': [{'i': 3, 's': 'd'}, {'i': 4, 's': 'e'}]},
        ]

        with self.create_table(columns, flatten_nested=0):
            self.client.execute(
                'INSERT INTO test (n) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                "[(0,'a'),(1,'b')]\n[(3,'d'),(4,'e')]\n"
            )

            inserted = self.client.execute(query)
            self.assertEqual(
                inserted,
                [([(0, 'a'), (1, 'b')],), ([(3, 'd'), (4, 'e')],)]
            )

    def test_get_nested_columns(self):
        self.assertEqual(
            nestedcolumn.get_nested_columns(
                'Nested(a Tuple(Array(Int8)),\n b Nullable(String))',
            ),
            ['Tuple(Array(Int8))', 'Nullable(String)']
        )

    def test_get_columns_with_types(self):
        self.assertEqual(
            nestedcolumn.get_columns_with_types(
                'Nested(a Tuple(Array(Int8)),\n b Nullable(String))',
            ),
            [('a', 'Tuple(Array(Int8))'), ('b', 'Nullable(String)')]
        )



from tests.testcase import BaseTestCase


class GeoTypesTestCase(BaseTestCase):
    required_server_version = (20, 5, 1)

    def client_kwargs(self, version):
        return {'settings': {'allow_experimental_geo_types': True}}

    def cli_client_kwargs(self):
        return {'allow_experimental_geo_types': 1}

    def entuple(self, lst):
        return tuple(
            self.entuple(x) if isinstance(x, list) else x for x in lst
        )

    def test_point(self):
        columns = 'a Point'
        data = [
            ((1.5, 2), ),
            ((3, 4), )
        ]

        with self.create_table(columns):
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '(1.5,2)\n(3,4)\n')

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_ring(self):
        columns = 'a Ring'
        data = [
            ([(1.5, 2), (3, 4)], )
        ]

        with self.create_table(columns):
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '[(1.5,2),(3,4)]\n')

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_polygon(self):
        columns = 'a Polygon'
        data = [
            ([
                [(1.5, 2), (3, 4)],
                [(5.5, 6), (7, 8)],
             ], )
        ]

        with self.create_table(columns):
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '[[(1.5,2),(3,4)],[(5.5,6),(7,8)]]\n')

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_multipolygon(self):
        columns = 'a MultiPolygon'
        data = [
            ([
                [
                    [(1.5, 2), (3, 4)],
                    [(5.5, 6), (7, 8)],
                ],
                [
                    [(2.5, 3), (4, 5)],
                    [(6.5, 7), (8, 9)],
                ]
             ], )
        ]

        with self.create_table(columns):
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                '['
                '[[(1.5,2),(3,4)],[(5.5,6),(7,8)]],'
                '[[(2.5,3),(4,5)],[(6.5,7),(8,9)]]'
                ']\n'
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)



import types
from collections import namedtuple
from contextlib import contextmanager
import socket
from unittest.mock import patch

from clickhouse_driver import connect
from clickhouse_driver.dbapi import (
    ProgrammingError, InterfaceError, OperationalError
)
from clickhouse_driver.dbapi.extras import DictCursor, NamedTupleCursor
from tests.testcase import BaseTestCase


class DBAPITestCaseBase(BaseTestCase):
    def create_connection(self, **kwargs):
        kwargs.setdefault('user', self.user)
        kwargs.setdefault('password', self.password)
        kwargs.setdefault('host', self.host)
        kwargs.setdefault('port', self.port)
        kwargs.setdefault('database', self.database)

        return connect(**kwargs)

    @contextmanager
    def created_connection(self, **kwargs):
        connection = self.create_connection(**kwargs)
        try:
            yield connection
        finally:
            connection.close()

    @contextmanager
    def created_cursor(self, **kwargs):
        cursor_kwargs = kwargs.pop('cursor_kwargs', {})
        with self.created_connection(**kwargs) as connection:
            cursor = connection.cursor(**cursor_kwargs)

            try:
                yield cursor
            finally:
                cursor.close()


class DBAPITestCase(DBAPITestCaseBase):
    def test_no_host_and_dsn(self):
        with self.assertRaises(ValueError) as e:
            self.create_connection(host=None)
            self.assertEqual(str(e.exception), 'host or dsn is required')

    def test_simple(self):
        with self.created_connection() as connection:
            cursor = connection.cursor()
            rv = cursor.execute('SELECT 1')
            self.assertIsNone(rv)
            self.assertEqual(cursor.fetchall(), [(1, )])

    def test_from_dsn(self):
        connection = connect(
            'clickhouse://{user}:{password}@{host}:{port}/{database}'.format(
                user=self.user, password=self.password,
                host=self.host, port=self.port, database=self.database
            )
        )
        cursor = connection.cursor()
        rv = cursor.execute('SELECT 1')
        self.assertIsNone(rv)
        self.assertEqual(cursor.fetchall(), [(1, )])
        connection.close()

    def test_connect_default_params(self):
        connection = connect(host=self.host)
        cursor = connection.cursor()

        rv = cursor.execute('SELECT 1')
        self.assertIsNone(rv)
        self.assertEqual(cursor.fetchall(), [(1, )])
        connection.close()

    def test_execute_fetchone(self):
        with self.created_cursor() as cursor:
            cursor.execute('SELECT number FROM system.numbers LIMIT 4')

            self.assertIsInstance(cursor._rows, list)

            self.assertEqual(cursor.fetchone(), (0, ))
            self.assertEqual(cursor.fetchone(), (1, ))
            self.assertEqual(cursor.fetchone(), (2, ))
            self.assertEqual(cursor.fetchone(), (3, ))
            self.assertEqual(cursor.fetchone(), None)

    def test_execute_fetchmany(self):
        with self.created_cursor() as cursor:
            cursor.execute('SELECT number FROM system.numbers LIMIT 4')

            self.assertIsInstance(cursor._rows, list)
            self.assertEqual(cursor.fetchmany(), [(0, )])
            self.assertEqual(cursor.fetchmany(None), [(1, )])
            self.assertEqual(cursor.fetchmany(0), [])
            self.assertEqual(cursor.fetchmany(-1), [(2, ), (3, )])

            cursor.execute('SELECT number FROM system.numbers LIMIT 4')
            self.assertEqual(cursor.fetchmany(1), [(0, )])
            self.assertEqual(cursor.fetchmany(2), [(1, ), (2, )])
            self.assertEqual(cursor.fetchmany(3), [(3, )])
            self.assertEqual(cursor.fetchmany(3), [])

            cursor.arraysize = 2
            cursor.execute('SELECT number FROM system.numbers LIMIT 4')
            self.assertEqual(cursor.fetchmany(), [(0,), (1,)])
            self.assertEqual(cursor.fetchmany(), [(2,), (3,)])

    def test_execute_fetchall(self):
        with self.created_cursor() as cursor:
            cursor.execute('SELECT number FROM system.numbers LIMIT 4')
            self.assertEqual(cursor.rowcount, 4)
            self.assertEqual(cursor.fetchall(), [(0,), (1,), (2,), (3,)])

    def test_executemany(self):
        with self.created_cursor() as cursor, self.create_table('a UInt32'):
            data = [(0, ), (1, ), (2, )]
            rv = cursor.executemany('INSERT INTO test VALUES', data)
            self.assertIsNone(rv, None)
            self.assertEqual(cursor.rowcount, 3)

            cursor.execute('SELECT * FROM test')
            self.assertEqual(cursor.fetchall(), data)

    def test_fake_transactions(self):
        with self.created_connection() as connection:
            connection.commit()
            connection.rollback()

    def test_exception_execute(self):
        with self.created_cursor(user='wrong_user') as cursor:
            with self.assertRaises(OperationalError) as e:
                cursor.execute('SELECT 1')
            code = 516 if self.server_version > (20, ) else 192
            self.assertIn('Code: {}'.format(code), str(e.exception))

    def test_exception_executemany(self):
        with self.created_cursor(user='wrong_user') as cursor:
            with self.assertRaises(OperationalError) as e:
                cursor.executemany('INSERT INTO test VALUES', [(0, )])
            code = 516 if self.server_version > (20, ) else 192
            self.assertIn('Code: {}'.format(code), str(e.exception))
            self.assertEqual(cursor.rowcount, -1)

    def test_rowcount_insert_from_select(self):
        with self.created_cursor() as cursor, self.create_table('a UInt8'):
            cursor.execute(
                'INSERT INTO test '
                'SELECT number FROM system.numbers LIMIT 4'
            )
            self.assertEqual(cursor.rowcount, -1)

    def test_execute_insert(self):
        with self.created_cursor() as cursor, self.create_table('a UInt8'):
            cursor.execute('INSERT INTO test VALUES', [[4]])
            self.assertEqual(cursor.rowcount, 1)

    def test_description(self):
        with self.created_cursor() as cursor:
            self.assertIsNone(cursor.description)
            cursor.execute('SELECT CAST(1 AS UInt32) AS test')
            desc = cursor.description
            self.assertEqual(len(desc), 1)
            self.assertEqual(desc[0].name, 'test')
            self.assertEqual(desc[0].type_code, 'UInt32')

    def test_pep249_sizes(self):
        with self.created_cursor() as cursor:
            cursor.setinputsizes(0)
            cursor.setoutputsize(0)

    def test_ddl(self):
        with self.created_cursor() as cursor:
            cursor.execute('DROP TABLE IF EXISTS test')
            self.assertEqual(cursor.fetchall(), [])
            self.assertEqual(cursor.rowcount, -1)

    def test_remember_last_host(self):
        self.n_calls = 0
        getaddrinfo = socket.getaddrinfo

        def side_getaddrinfo(host, *args, **kwargs):
            if host == 'wrong_host':
                self.n_calls += 1
                raise socket.error(-2, 'Name or service not known')
            return getaddrinfo(host, *args, **kwargs)

        with patch('socket.getaddrinfo') as mocked_getaddrinfo:
            mocked_getaddrinfo.side_effect = side_getaddrinfo

            conn_kwargs = {
                'host': 'wrong_host',
                'port': 1234,
                'alt_hosts': '{}:{}'.format(self.host, self.port)
            }
            with self.created_connection(**conn_kwargs) as connection:
                cursor = connection.cursor()
                cursor.execute('SELECT 1')
                self.assertEqual(cursor.fetchall(), [(1, )])
                cursor.close()

                cursor = connection.cursor()
                cursor.execute('SELECT 1')
                self.assertEqual(cursor.fetchall(), [(1, )])
                cursor.close()

        # Last host must be remembered and getaddrinfo must call exactly
        # once with host == 'wrong_host'.
        self.assertEqual(self.n_calls, 1)

    def test_remove_cursor_from_connection_on_closing(self):
        with self.created_connection() as connection:
            self.assertEqual(len(connection.cursors), 0)
            cur = connection.cursor()
            self.assertEqual(len(connection.cursors), 1)
            cur.close()
            self.assertEqual(len(connection.cursors), 0)


class StreamingTestCase(DBAPITestCaseBase):
    def test_fetchone(self):
        with self.created_cursor() as cursor:
            cursor.set_stream_results(True, 2)
            cursor.execute('SELECT number FROM system.numbers LIMIT 4')

            self.assertIsInstance(cursor._rows, types.GeneratorType)

            self.assertEqual(cursor.fetchone(), (0, ))
            self.assertEqual(cursor.fetchone(), (1, ))
            self.assertEqual(cursor.fetchone(), (2, ))
            self.assertEqual(cursor.fetchone(), (3, ))
            self.assertEqual(cursor.fetchone(), None)

    def test_fetchmany(self):
        with self.created_cursor() as cursor:
            cursor.set_stream_results(True, 2)
            cursor.execute('SELECT number FROM system.numbers LIMIT 4')

            self.assertIsInstance(cursor._rows, types.GeneratorType)
            self.assertEqual(cursor.fetchmany(), [(0, )])
            self.assertEqual(cursor.fetchmany(None), [(1, )])
            self.assertEqual(cursor.fetchmany(0), [])
            self.assertEqual(cursor.fetchmany(-1), [(2, ), (3, )])

            cursor.execute('SELECT number FROM system.numbers LIMIT 4')
            self.assertEqual(cursor.fetchmany(1), [(0, )])
            self.assertEqual(cursor.fetchmany(2), [(1, ), (2, )])
            self.assertEqual(cursor.fetchmany(3), [(3, )])
            self.assertEqual(cursor.fetchmany(3), [])

    def test_fetchall(self):
        with self.created_cursor() as cursor:
            cursor.set_stream_results(True, 2)

            cursor.execute('SELECT number FROM system.numbers LIMIT 4')
            # Check rowcount before and after fetch.
            self.assertEqual(cursor.rowcount, -1)
            self.assertEqual(cursor.fetchall(), [(0, ), (1, ), (2, ), (3, )])
            self.assertEqual(cursor.rowcount, -1)


class ReprTestCase(DBAPITestCaseBase):
    def test_cursor_repr(self):
        with self.created_cursor() as cursor:
            self.assertIn('closed: False', str(cursor))

    def test_connection_repr(self):
        with self.created_connection() as conn:
            self.assertIn('closed: False', str(conn))


class ExtrasTestCase(DBAPITestCaseBase):
    def test_columns_with_types_select(self):
        with self.created_cursor() as cursor:
            self.assertIsNone(cursor.columns_with_types)
            cursor.execute(
                'SELECT CAST(number AS UInt64) AS x '
                'FROM system.numbers LIMIT 4'
            )
            cursor.fetchall()
            self.assertEqual(cursor.columns_with_types, [('x', 'UInt64')])

    def test_columns_with_types_insert(self):
        with self.created_cursor() as cursor, self.create_table('a UInt8'):
            cursor.executemany('INSERT INTO test (a) VALUES', [(123, )])
            self.assertIsNone(cursor.columns_with_types)

    def test_columns_with_types_streaming(self):
        with self.created_cursor() as cursor:
            cursor.set_stream_results(True, 2)
            cursor.execute(
                'SELECT CAST(number AS UInt64) AS x '
                'FROM system.numbers LIMIT 4'
            )
            self.assertEqual(cursor.columns_with_types, [('x', 'UInt64')])
            list(cursor)
            self.assertEqual(cursor.columns_with_types, [('x', 'UInt64')])

    def test_set_external_tables(self):
        with self.created_cursor() as cursor:
            data = [(0, ), (1, ), (2, )]
            cursor.set_external_table('table1', [('x', 'UInt32')], data)
            cursor.execute('SELECT * FROM table1')
            self.assertEqual(cursor.fetchall(), data)

    def test_settings(self):
        with self.created_cursor() as cursor:
            cursor.set_settings({'max_threads': 100500})

            cursor.execute(
                "SELECT name, value, changed FROM system.settings "
                "WHERE name = 'max_threads'",
            )
            self.assertEqual(cursor.fetchall(), [('max_threads', '100500', 1)])

    def test_set_query_id(self):
        with self.created_cursor() as cursor:
            query_id = 'my_query_id'
            cursor.set_query_id(query_id)
            cursor.execute(
                'SELECT query_id '
                'FROM system.processes '
                'WHERE query_id = %(query_id)s',
                {'query_id': query_id}
            )
            self.assertEqual(cursor.fetchall(), [(query_id, )])

    def test_types_check(self):
        with self.created_cursor() as cursor, self.create_table('a UInt8'):
            cursor.set_types_check(True)

            data = [(300, )]
            cursor.executemany('INSERT INTO test (a) VALUES', data)
            cursor.execute('SELECT * FROM test')
            self.assertEqual(cursor.fetchall(), [(44, )])

    def test_cursor_iteration(self):
        with self.created_cursor() as cursor:
            cursor.execute('SELECT number FROM system.numbers LIMIT 4')
            self.assertEqual(list(cursor), [(0,), (1,), (2,), (3,)])

    def test_context_managers(self):
        with self.create_connection() as connection:
            with connection.cursor() as cursor:
                cursor.execute('SELECT 1')
                self.assertEqual(cursor.fetchall(), [(1, )])


class InterfaceTestCase(DBAPITestCaseBase):
    def test_execute_after_close(self):
        with self.created_cursor() as cursor:
            cursor.close()
            with self.assertRaises(InterfaceError) as e:
                cursor.execute('SELECT 1')
            self.assertEqual(str(e.exception), 'cursor already closed')

    def test_create_cursor_on_closed_connection(self):
        connection = self.create_connection()
        connection.close()

        with self.assertRaises(InterfaceError) as e:
            connection.cursor()
            self.assertEqual(str(e.exception), 'connection already closed')

    def test_execute_fetch_before_query(self):
        with self.created_cursor() as cursor:
            with self.assertRaises(ProgrammingError) as e:
                cursor.fetchall()
            self.assertEqual(str(e.exception), 'no results to fetch')


class DictCursorFactoryTestCase(DBAPITestCaseBase):
    def test_execute_fetchone(self):
        cursor_kwargs = {'cursor_factory': DictCursor}

        with self.created_cursor(cursor_kwargs=cursor_kwargs) as cursor:
            cursor.execute('SELECT number FROM system.numbers LIMIT 4')

            self.assertIsInstance(cursor._rows, list)
            self.assertEqual(cursor.fetchone(), {'number': 0})
            self.assertEqual(cursor.fetchone(), {'number': 1})
            self.assertEqual(cursor.fetchone(), {'number': 2})
            self.assertEqual(cursor.fetchone(), {'number': 3})
            self.assertEqual(cursor.fetchone(), None)

    def test_execute_fetchmany(self):
        cursor_kwargs = {'cursor_factory': DictCursor}

        with self.created_cursor(cursor_kwargs=cursor_kwargs) as cursor:
            cursor.execute('SELECT number FROM system.numbers LIMIT 4')

            self.assertIsInstance(cursor._rows, list)
            self.assertEqual(cursor.fetchmany(), [{'number': 0}])
            self.assertEqual(cursor.fetchmany(None), [{'number': 1}])
            self.assertEqual(cursor.fetchmany(0), [])
            self.assertEqual(
                cursor.fetchmany(-1), [{'number': 2}, {'number': 3}]
            )

            cursor.execute('SELECT number FROM system.numbers LIMIT 4')
            self.assertEqual(cursor.fetchmany(1), [{'number': 0}])
            self.assertEqual(
                cursor.fetchmany(2), [{'number': 1}, {'number': 2}]
            )
            self.assertEqual(cursor.fetchmany(3), [{'number': 3}])
            self.assertEqual(cursor.fetchmany(3), [])

            cursor.arraysize = 2
            cursor.execute('SELECT number FROM system.numbers LIMIT 4')
            self.assertEqual(
                cursor.fetchmany(), [{'number': 0}, {'number': 1}])
            self.assertEqual(
                cursor.fetchmany(), [{'number': 2}, {'number': 3}]
            )

    def test_execute_fetchall(self):
        cursor_kwargs = {'cursor_factory': DictCursor}

        with self.created_cursor(cursor_kwargs=cursor_kwargs) as cursor:
            cursor.execute('SELECT number FROM system.numbers LIMIT 4')
            self.assertEqual(cursor.rowcount, 4)
            self.assertEqual(
                cursor.fetchall(), [
                    {'number': 0}, {'number': 1}, {'number': 2}, {'number': 3}
                ])


class NamedTupleCursorFactoryTestCase(DBAPITestCaseBase):
    def test_execute_fetchone(self):
        cursor_kwargs = {'cursor_factory': NamedTupleCursor}

        with self.created_cursor(cursor_kwargs=cursor_kwargs) as cursor:
            cursor.execute('SELECT number FROM system.numbers LIMIT 4')

            self.assertIsInstance(cursor._rows, list)
            nt = namedtuple('Record', cursor._columns)

            self.assertEqual(cursor.fetchone(), nt(0))
            self.assertEqual(cursor.fetchone(), nt(1))
            self.assertEqual(cursor.fetchone(), nt(2))
            self.assertEqual(cursor.fetchone(), nt(3))
            self.assertEqual(cursor.fetchone(), None)

    def test_execute_fetchmany(self):
        cursor_kwargs = {'cursor_factory': NamedTupleCursor}

        with self.created_cursor(cursor_kwargs=cursor_kwargs) as cursor:
            cursor.execute('SELECT number FROM system.numbers LIMIT 4')

            self.assertIsInstance(cursor._rows, list)
            nt = namedtuple('Record', cursor._columns)

            self.assertEqual(cursor.fetchmany(), [nt(0)])
            self.assertEqual(cursor.fetchmany(None), [nt(1)])
            self.assertEqual(cursor.fetchmany(0), [])
            self.assertEqual(cursor.fetchmany(-1), [nt(2), nt(3)])

            cursor.execute('SELECT number FROM system.numbers LIMIT 4')
            self.assertEqual(cursor.fetchmany(1), [nt(0)])
            self.assertEqual(cursor.fetchmany(2), [nt(1), nt(2)])
            self.assertEqual(cursor.fetchmany(3), [nt(3)])
            self.assertEqual(cursor.fetchmany(3), [])

            cursor.arraysize = 2
            cursor.execute('SELECT number FROM system.numbers LIMIT 4')
            self.assertEqual(cursor.fetchmany(), [nt(0), nt(1)])
            self.assertEqual(cursor.fetchmany(), [nt(2), nt(3)])

    def test_execute_fetchall(self):
        cursor_kwargs = {'cursor_factory': NamedTupleCursor}

        with self.created_cursor(cursor_kwargs=cursor_kwargs) as cursor:
            cursor.execute('SELECT number FROM system.numbers LIMIT 4')
            self.assertEqual(cursor.rowcount, 4)
            nt = namedtuple('Record', cursor._columns)

            self.assertEqual(cursor.fetchall(), [nt(0), nt(1), nt(2), nt(3)])

    def test_make_nt_caching(self):
        cursor_kwargs = {'cursor_factory': NamedTupleCursor}

        with self.created_cursor(cursor_kwargs=cursor_kwargs) as cursor:
            cursor.execute('SELECT number FROM system.numbers LIMIT 4')

            self.assertIsInstance(cursor._rows, list)
            nt = namedtuple('Record', cursor._columns)

            self.assertEqual(cursor.fetchone(), nt(0))

            with patch('clickhouse_driver.dbapi.extras.namedtuple') as nt_mock:
                nt_mock.side_effect = ValueError
                self.assertEqual(cursor.fetchone(), nt(1))



from tests.testcase import BaseTestCase


class CommonTestCase(BaseTestCase):
    client_kwargs = {'settings': {'insert_block_size': 1}}

    def setUp(self):
        super(CommonTestCase, self).setUp()

        self.send_data_count = 0
        old_send_data = self.client.connection.send_data

        def send_data(*args, **kwargs):
            self.send_data_count += 1
            return old_send_data(*args, **kwargs)

        self.client.connection.send_data = send_data

    def test_insert_block_size(self):
        with self.create_table('a UInt8'):
            data = [(x, ) for x in range(4)]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )
            # Two empty blocks: for end of sending external tables
            # and data.
            self.assertEqual(self.send_data_count, 4 + 2)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '0\n1\n2\n3\n')
            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_columnar_insert_block_size(self):
        with self.create_table('a UInt8'):
            data = [(0, 1, 2, 3)]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, columnar=True
            )
            # Two empty blocks: for end of sending external tables
            # and data.
            self.assertEqual(self.send_data_count, 4 + 2)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '0\n1\n2\n3\n')
            inserted = self.client.execute(query)
            expected = [(0, ), (1, ), (2, ), (3, )]
            self.assertEqual(inserted, expected)




from tests.testcase import BaseTestCase


class ExternalTablesTestCase(BaseTestCase):
    def test_select(self):
        tables = [{
            'name': 'test',
            'structure': [('x', 'Int32'), ('y', 'Array(Int32)')],
            'data': [
                {'x': 100, 'y': [2, 4, 6, 8]},
                {'x': 500, 'y': [1, 3, 5, 7]},
            ]
        }]
        rv = self.client.execute('SELECT * FROM test', external_tables=tables)
        self.assertEqual(rv, [(100, [2, 4, 6, 8]), (500, [1, 3, 5, 7])])

    def test_send_empty_table(self):
        tables = [{
            'name': 'test',
            'structure': [('x', 'Int32')],
            'data': []
        }]
        rv = self.client.execute('SELECT * FROM test', external_tables=tables)
        self.assertEqual(rv, [])

    def test_send_empty_table_structure(self):
        tables = [{
            'name': 'test',
            'structure': [],
            'data': []
        }]
        with self.assertRaises(ValueError) as e:
            self.client.execute('SELECT * FROM test', external_tables=tables)

        self.assertIn('Empty table "test" structure', str(e.exception))



try:
    import numpy as np
except ImportError:
    np = None

try:
    import pandas as pd
except ImportError:
    pd = None


from tests.numpy.testcase import NumpyBaseTestCase

# from decimal import Decimal


class LowCardinalityTestCase(NumpyBaseTestCase):
    required_server_version = (19, 3, 3)
    stable_support_version = (19, 9, 2)

    def cli_client_kwargs(self):
        if self.server_version >= self.stable_support_version:
            return {'allow_suspicious_low_cardinality_types': 1}

    def check_result(self, inserted, data):
        self.assertArraysEqual(inserted[0], data[0])
        self.assertIsInstance(inserted[0], pd.Categorical)

    def test_uint8(self):
        with self.create_table('a LowCardinality(UInt8)'):
            data = [np.array(range(255))]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, columnar=True
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                '\n'.join(str(x) for x in data[0]) + '\n'
            )

            inserted = self.client.execute(query, columnar=True)
            self.check_result(inserted, data)

    def test_int8(self):
        with self.create_table('a LowCardinality(Int8)'):
            data = [np.array([x - 127 for x in range(255)])]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, columnar=True
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                '\n'.join(str(x) for x in data[0]) + '\n'

            )

            inserted = self.client.execute(query, columnar=True)
            self.check_result(inserted, data)

    def test_nullable_int8(self):
        with self.create_table('a LowCardinality(Nullable(Int8))'):
            data = [np.array([None, -1, 0, 1, None], dtype=object)]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, columnar=True
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '\\N\n-1\n0\n1\n\\N\n')

            inserted = self.client.execute(query, columnar=True)
            self.assertArraysEqual(
                inserted[0].astype(str),
                pd.Categorical(data[0]).astype(str)
            )
            self.assertIsInstance(inserted[0], pd.Categorical)

    def test_date(self):
        with self.create_table('a LowCardinality(Date)'):
            data = [np.array(list(range(300)), dtype='datetime64[D]')]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, columnar=True
            )

            query = 'SELECT * FROM test'
            inserted = self.client.execute(query, columnar=True)
            self.check_result(inserted, data)

    def test_float(self):
        with self.create_table('a LowCardinality(Float)'):
            data = [np.array([float(x) for x in range(300)])]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, columnar=True
            )

            query = 'SELECT * FROM test'
            inserted = self.client.execute(query, columnar=True)
            self.check_result(inserted, data)

    # def test_decimal(self):
    #     with self.create_table('a LowCardinality(Float)'):
    #         data = [(Decimal(x),) for x in range(300)]
    #         self.client.execute('INSERT INTO test (a) VALUES', data[0])
    #
    #         query = 'SELECT * FROM test'
    #         inserted = self.client.execute(query)
    #         self.assertEqual(inserted, data[0])
    #
    # def test_array(self):
    #     with self.create_table('a Array(LowCardinality(Int16))'):
    #         data = [([100, 500], )]
    #         self.client.execute('INSERT INTO test (a) VALUES', data[0])
    #
    #         query = 'SELECT * FROM test'
    #         inserted = self.emit_cli(query)
    #         self.assertEqual(inserted, '[100,500]\n')
    #
    #         inserted = self.client.execute(query)
    #         self.assertEqual(inserted, data[0])
    #
    # def test_empty_array(self):
    #     with self.create_table('a Array(LowCardinality(Int16))'):
    #         data = [([], )]
    #         self.client.execute('INSERT INTO test (a) VALUES', data[0])
    #
    #         query = 'SELECT * FROM test'
    #         inserted = self.emit_cli(query)
    #         self.assertEqual(inserted, '[]\n')
    #
    #         inserted = self.client.execute(query)
    #         self.assertEqual(inserted, data[0])
    #
    def test_string(self):
        with self.create_table('a LowCardinality(String)'):
            data = [
                np.array(['test', 'low', 'cardinality', 'test', 'test', ''])
            ]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, columnar=True
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                'test\nlow\ncardinality\ntest\ntest\n\n'
            )

            inserted = self.client.execute(query, columnar=True)
            self.check_result(inserted, data)

    def test_insert_nan_string_into_non_nullable(self):
        with self.create_table('a LowCardinality(String)'):
            data = [
                np.array(['test', None], dtype=object)
            ]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, columnar=True
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                'test\n\n'
            )

            inserted = self.client.execute(query, columnar=True)
            self.check_result(inserted, [np.array(['test', ''])])

    def test_fixed_string(self):
        with self.create_table('a LowCardinality(FixedString(12))'):
            data = [
                np.array(['test', 'low', 'cardinality', 'test', 'test', ''])
            ]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, columnar=True
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                'test\\0\\0\\0\\0\\0\\0\\0\\0\n'
                'low\\0\\0\\0\\0\\0\\0\\0\\0\\0\n'
                'cardinality\\0\n'
                'test\\0\\0\\0\\0\\0\\0\\0\\0\n'
                'test\\0\\0\\0\\0\\0\\0\\0\\0\n'
                '\\0\\0\\0\\0\\0\\0\\0\\0\\0\\0\\0\\0\n'
            )

            inserted = self.client.execute(query, columnar=True)
            self.check_result(inserted, data)

    def test_nullable_string(self):
        with self.create_table('a LowCardinality(Nullable(String))'):
            data = [
                np.array(['test', '', None], dtype=object)
            ]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, columnar=True
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                'test\n\n\\N\n'
            )

            inserted = self.client.execute(query, columnar=True)
            self.assertArraysEqual(
                inserted[0].astype(str), pd.Categorical(data[0]).astype(str)
            )
            self.assertIsInstance(inserted[0], pd.Categorical)



import ssl

from clickhouse_driver import Client
from clickhouse_driver.compression.lz4 import Compressor as LZ4Compressor
from clickhouse_driver.compression.lz4hc import Compressor as LZHC4Compressor
from clickhouse_driver.compression.zstd import Compressor as ZSTDCompressor
from clickhouse_driver.protocol import Compression
from tests.numpy.util import check_numpy
from tests.testcase import TestCase


class ClientFromUrlTestCase(TestCase):
    def assertHostsEqual(self, client, another, msg=None):
        self.assertEqual(list(client.connection.hosts), another, msg=msg)

    def test_simple(self):
        c = Client.from_url('clickhouse://host')

        self.assertHostsEqual(c, [('host', 9000)])
        self.assertEqual(c.connection.database, '')

        c = Client.from_url('clickhouse://host/db')

        self.assertHostsEqual(c, [('host', 9000)])
        self.assertEqual(c.connection.database, 'db')

    def test_credentials(self):
        c = Client.from_url('clickhouse://host/db')

        self.assertEqual(c.connection.user, 'default')
        self.assertEqual(c.connection.password, '')

        c = Client.from_url('clickhouse://admin:secure@host/db')

        self.assertEqual(c.connection.user, 'admin')
        self.assertEqual(c.connection.password, 'secure')

        c = Client.from_url('clickhouse://user:@host/db')

        self.assertEqual(c.connection.user, 'user')
        self.assertEqual(c.connection.password, '')

    def test_credentials_unquoting(self):
        c = Client.from_url('clickhouse://ad%3Amin:se%2Fcure@host/db')

        self.assertEqual(c.connection.user, 'ad:min')
        self.assertEqual(c.connection.password, 'se/cure')

    def test_schema(self):
        c = Client.from_url('clickhouse://host')
        self.assertFalse(c.connection.secure_socket)

        c = Client.from_url('clickhouses://host')
        self.assertTrue(c.connection.secure_socket)

        c = Client.from_url('test://host')
        self.assertFalse(c.connection.secure_socket)

    def test_port(self):
        c = Client.from_url('clickhouse://host')
        self.assertHostsEqual(c, [('host', 9000)])

        c = Client.from_url('clickhouses://host')
        self.assertHostsEqual(c, [('host', 9440)])

        c = Client.from_url('clickhouses://host:1234')
        self.assertHostsEqual(c, [('host', 1234)])

    def test_secure(self):
        c = Client.from_url('clickhouse://host?secure=n')
        self.assertHostsEqual(c, [('host', 9000)])
        self.assertFalse(c.connection.secure_socket)

        c = Client.from_url('clickhouse://host?secure=y')
        self.assertHostsEqual(c, [('host', 9440)])
        self.assertTrue(c.connection.secure_socket)

        c = Client.from_url('clickhouse://host:1234?secure=y')
        self.assertHostsEqual(c, [('host', 1234)])
        self.assertTrue(c.connection.secure_socket)

        with self.assertRaises(ValueError):
            Client.from_url('clickhouse://host:1234?secure=nonono')

    def test_compression(self):
        c = Client.from_url('clickhouse://host?compression=n')
        self.assertEqual(c.connection.compression, Compression.DISABLED)
        self.assertIsNone(c.connection.compressor_cls)

        c = Client.from_url('clickhouse://host?compression=y')
        self.assertEqual(c.connection.compression, Compression.ENABLED)
        self.assertIs(c.connection.compressor_cls, LZ4Compressor)

        c = Client.from_url('clickhouse://host?compression=lz4')
        self.assertEqual(c.connection.compression, Compression.ENABLED)
        self.assertIs(c.connection.compressor_cls, LZ4Compressor)

        c = Client.from_url('clickhouse://host?compression=lz4hc')
        self.assertEqual(c.connection.compression, Compression.ENABLED)
        self.assertIs(c.connection.compressor_cls, LZHC4Compressor)

        c = Client.from_url('clickhouse://host?compression=zstd')
        self.assertEqual(c.connection.compression, Compression.ENABLED)
        self.assertIs(c.connection.compressor_cls, ZSTDCompressor)

        with self.assertRaises(ValueError):
            Client.from_url('clickhouse://host:1234?compression=custom')

    def test_client_name(self):
        c = Client.from_url('clickhouse://host?client_name=native')
        self.assertEqual(c.connection.client_name, 'ClickHouse native')

    def test_timeouts(self):
        with self.assertRaises(ValueError):
            Client.from_url('clickhouse://host?connect_timeout=test')

        c = Client.from_url('clickhouse://host?connect_timeout=1.2')
        self.assertEqual(c.connection.connect_timeout, 1.2)

        c = Client.from_url('clickhouse://host?send_receive_timeout=1.2')
        self.assertEqual(c.connection.send_receive_timeout, 1.2)

        c = Client.from_url('clickhouse://host?sync_request_timeout=1.2')
        self.assertEqual(c.connection.sync_request_timeout, 1.2)

    def test_compress_block_size(self):
        with self.assertRaises(ValueError):
            Client.from_url('clickhouse://host?compress_block_size=test')

        c = Client.from_url('clickhouse://host?compress_block_size=100500')
        # compression is not set
        self.assertIsNone(c.connection.compress_block_size)

        c = Client.from_url(
            'clickhouse://host?'
            'compress_block_size=100500&'
            'compression=1'
        )
        self.assertEqual(c.connection.compress_block_size, 100500)

    def test_settings(self):
        c = Client.from_url(
            'clickhouse://host?'
            'send_logs_level=trace&'
            'max_block_size=123'
        )
        self.assertEqual(c.settings, {
            'send_logs_level': 'trace',
            'max_block_size': '123'
        })

    def test_ssl(self):
        c = Client.from_url(
            'clickhouses://host?'
            'verify=false&'
            'ssl_version=PROTOCOL_SSLv23&'
            'ca_certs=/tmp/certs&'
            'ciphers=HIGH:-aNULL:-eNULL:-PSK:RC4-SHA:RC4-MD5'
        )
        self.assertEqual(c.connection.ssl_options, {
            'ssl_version': ssl.PROTOCOL_SSLv23,
            'ca_certs': '/tmp/certs',
            'ciphers': 'HIGH:-aNULL:-eNULL:-PSK:RC4-SHA:RC4-MD5'
        })

    def test_ssl_key_cert(self):
        base_url = (
            'clickhouses://host?'
            'verify=true&'
            'ssl_version=PROTOCOL_SSLv23&'
            'ca_certs=/tmp/certs&'
            'ciphers=HIGH:-aNULL:-eNULL:-PSK:RC4-SHA:RC4-MD5&'
        )
        base_expected = {
            'ssl_version': ssl.PROTOCOL_SSLv23,
            'ca_certs': '/tmp/certs',
            'ciphers': 'HIGH:-aNULL:-eNULL:-PSK:RC4-SHA:RC4-MD5'
        }

        c = Client.from_url(
            base_url +
            'keyfile=/tmp/client.key&'
            'certfile=/tmp/client.cert'
        )
        expected = base_expected.copy()
        expected.update({
            'keyfile': '/tmp/client.key',
            'certfile': '/tmp/client.cert'
        })
        self.assertEqual(c.connection.ssl_options, expected)

        c = Client.from_url(
            base_url +
            'certfile=/tmp/client.cert'
        )
        expected = base_expected.copy()
        expected.update({
            'certfile': '/tmp/client.cert'
        })
        self.assertEqual(c.connection.ssl_options, expected)

    def test_alt_hosts(self):
        c = Client.from_url('clickhouse://host?alt_hosts=host2:1234')
        self.assertHostsEqual(c, [('host', 9000), ('host2', 1234)])

        c = Client.from_url('clickhouse://host?alt_hosts=host2')
        self.assertHostsEqual(c, [('host', 9000), ('host2', 9000)])

    def test_parameters_cast(self):
        c = Client.from_url('clickhouse://host?insert_block_size=123')
        self.assertEqual(
            c.connection.context.client_settings['insert_block_size'], 123
        )

    def test_settings_is_important(self):
        c = Client.from_url('clickhouse://host?settings_is_important=1')
        self.assertEqual(c.connection.settings_is_important, True)

        with self.assertRaises(ValueError):
            c = Client.from_url('clickhouse://host?settings_is_important=2')
            self.assertEqual(c.connection.settings_is_important, True)

        c = Client.from_url('clickhouse://host?settings_is_important=0')
        self.assertEqual(c.connection.settings_is_important, False)

    @check_numpy
    def test_use_numpy(self):
        c = Client.from_url('clickhouse://host?use_numpy=true')
        self.assertTrue(c.connection.context.client_settings['use_numpy'])

    def test_opentelemetry(self):
        c = Client.from_url(
            'clickhouse://host?opentelemetry_traceparent='
            '00-aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa-bbbbbbbbbbbbbbbb-00'
        )
        self.assertEqual(
            c.connection.context.client_settings['opentelemetry_traceparent'],
            '00-aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa-bbbbbbbbbbbbbbbb-00'
        )
        self.assertEqual(
            c.connection.context.client_settings['opentelemetry_tracestate'],
            ''
        )

        c = Client.from_url(
            'clickhouse://host?opentelemetry_traceparent='
            '00-aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa-bbbbbbbbbbbbbbbb-00&'
            'opentelemetry_tracestate=state'
        )
        self.assertEqual(
            c.connection.context.client_settings['opentelemetry_traceparent'],
            '00-aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa-bbbbbbbbbbbbbbbb-00'
        )
        self.assertEqual(
            c.connection.context.client_settings['opentelemetry_tracestate'],
            'state'
        )

    def test_quota_key(self):
        c = Client.from_url('clickhouse://host?quota_key=myquota')
        self.assertEqual(
            c.connection.context.client_settings['quota_key'], 'myquota'
        )

        c = Client.from_url('clickhouse://host')
        self.assertEqual(
            c.connection.context.client_settings['quota_key'], ''
        )

    def test_round_robin(self):
        c = Client.from_url('clickhouse://host?alt_hosts=host2')
        self.assertEqual(len(c.connections), 0)

        c = Client.from_url(
            'clickhouse://host?round_robin=true&alt_hosts=host2'
        )
        self.assertEqual(len(c.connections), 1)



from cpython cimport Py_INCREF
from cpython.tuple cimport PyTuple_New, PyTuple_SET_ITEM

from .. import writer

cdef object MAX_UINT64 = writer.MAX_UINT64
cdef object MAX_INT64 = writer.MAX_INT64


def int128_from_quads(quad_items, unsigned long long n_items):
    cdef unsigned int factor = 2
    items = PyTuple_New(n_items)

    cdef unsigned long long i, ix
    cdef object item

    for i in range(n_items):
        ix = factor * i

        if quad_items[ix + 1] > MAX_INT64:
            item = (
                -((MAX_UINT64 - quad_items[ix + 1]) << 64)
                - (MAX_UINT64 - quad_items[ix])
                - 1
            )

        else:
            item = (quad_items[ix + 1] << 64) + quad_items[ix]

        Py_INCREF(item)
        PyTuple_SET_ITEM(items, i, item)

    return items


def int128_to_quads(items, unsigned long long n_items):
    cdef unsigned int factor = 2
    quad_items = PyTuple_New(n_items * factor)

    cdef unsigned long long i, ix
    cdef object x, item

    for i in range(n_items):
        ix = factor * i

        x = items[i]
        if x < 0:
            x = -x - 1

            item = MAX_UINT64 - x & MAX_UINT64
            Py_INCREF(item)
            PyTuple_SET_ITEM(quad_items, ix, item)

            item = MAX_UINT64 - (x >> 64) & MAX_UINT64
            Py_INCREF(item)
            PyTuple_SET_ITEM(quad_items, ix + 1, item)

        else:
            item = x & MAX_UINT64
            Py_INCREF(item)
            PyTuple_SET_ITEM(quad_items, ix, item)

            item = (x >> 64) & MAX_UINT64
            Py_INCREF(item)
            PyTuple_SET_ITEM(quad_items, ix + 1, item)

    return quad_items


def uint128_from_quads(quad_items, unsigned long long n_items):
    cdef unsigned int factor = 2
    items = PyTuple_New(n_items)

    cdef unsigned long long i, ix
    cdef object item

    for i in range(n_items):
        ix = factor * i
        item = (quad_items[ix + 1] << 64) + quad_items[ix]

        Py_INCREF(item)
        PyTuple_SET_ITEM(items, i, item)

    return items


def uint128_to_quads(items, unsigned long long n_items):
    cdef unsigned int factor = 2
    quad_items = PyTuple_New(n_items * factor)

    cdef unsigned long long i, ix
    cdef object x, item

    for i in range(n_items):
        ix = factor * i

        x = items[i]
        item = x & MAX_UINT64
        Py_INCREF(item)
        PyTuple_SET_ITEM(quad_items, ix, item)

        item = (x >> 64) & MAX_UINT64
        Py_INCREF(item)
        PyTuple_SET_ITEM(quad_items, ix + 1, item)

    return quad_items

# 256 bits


def int256_from_quads(quad_items, unsigned long long n_items):
    cdef unsigned int factor = 4
    items = PyTuple_New(n_items)

    cdef unsigned long long i, ix
    cdef object item

    for i in range(n_items):
        ix = factor * i

        if quad_items[ix + 3] > MAX_INT64:
            item = (
                -((MAX_UINT64 - quad_items[ix + 3]) << 192)
                -((MAX_UINT64 - quad_items[ix + 2]) << 128)
                -((MAX_UINT64 - quad_items[ix + 1]) << 64)
                - (MAX_UINT64 - quad_items[ix])
                - 1
            )

        else:
            item = (
                (quad_items[ix + 3] << 192) +
                (quad_items[ix + 2] << 128) +
                (quad_items[ix + 1] << 64) +
                quad_items[ix]
            )

        Py_INCREF(item)
        PyTuple_SET_ITEM(items, i, item)

    return items


def int256_to_quads(items, unsigned long long n_items):
    cdef unsigned int factor = 4
    quad_items = PyTuple_New(n_items * factor)

    cdef unsigned long long i, ix
    cdef object x, item

    for i in range(n_items):
        ix = factor * i

        x = items[i]
        if x < 0:
            x = -x - 1

            item = MAX_UINT64 - x & MAX_UINT64
            Py_INCREF(item)
            PyTuple_SET_ITEM(quad_items, ix, item)

            item = MAX_UINT64 - (x >> 64) & MAX_UINT64
            Py_INCREF(item)
            PyTuple_SET_ITEM(quad_items, ix + 1, item)

            item = MAX_UINT64 - (x >> 128) & MAX_UINT64
            Py_INCREF(item)
            PyTuple_SET_ITEM(quad_items, ix + 2, item)

            item = MAX_UINT64 - (x >> 192) & MAX_UINT64
            Py_INCREF(item)
            PyTuple_SET_ITEM(quad_items, ix + 3, item)

        else:
            item = x & MAX_UINT64
            Py_INCREF(item)
            PyTuple_SET_ITEM(quad_items, ix, item)

            item = (x >> 64) & MAX_UINT64
            Py_INCREF(item)
            PyTuple_SET_ITEM(quad_items, ix + 1, item)

            item = (x >> 128) & MAX_UINT64
            Py_INCREF(item)
            PyTuple_SET_ITEM(quad_items, ix + 2, item)

            item = (x >> 192) & MAX_UINT64
            Py_INCREF(item)
            PyTuple_SET_ITEM(quad_items, ix + 3, item)

    return quad_items


def uint256_from_quads(quad_items, unsigned long long n_items):
    cdef unsigned int factor = 4
    items = PyTuple_New(n_items)

    cdef unsigned long long i, ix
    cdef object item

    for i in range(n_items):
        ix = factor * i
        item = (
            (quad_items[ix + 3] << 192) +
            (quad_items[ix + 2] << 128) +
            (quad_items[ix + 1] << 64) +
            quad_items[ix]
        )

        Py_INCREF(item)
        PyTuple_SET_ITEM(items, i, item)

    return items


def uint256_to_quads(items, unsigned long long n_items):
    cdef unsigned int factor = 4
    quad_items = PyTuple_New(n_items * factor)

    cdef unsigned long long i, ix
    cdef object x, item

    for i in range(n_items):
        ix = factor * i

        x = items[i]
        item = x & MAX_UINT64
        Py_INCREF(item)
        PyTuple_SET_ITEM(quad_items, ix, item)

        item = (x >> 64) & MAX_UINT64
        Py_INCREF(item)
        PyTuple_SET_ITEM(quad_items, ix + 1, item)

        item = (x >> 128) & MAX_UINT64
        Py_INCREF(item)
        PyTuple_SET_ITEM(quad_items, ix + 2, item)

        item = (x >> 192) & MAX_UINT64
        Py_INCREF(item)
        PyTuple_SET_ITEM(quad_items, ix + 3, item)

    return quad_items



from datetime import date

from clickhouse_driver import errors
from tests.testcase import BaseTestCase
from tests.util import require_server_version


class TupleTestCase(BaseTestCase):
    def entuple(self, lst):
        return tuple(
            self.entuple(x) if isinstance(x, list) else x for x in lst
        )

    def test_simple(self):
        columns = 'a Tuple(Int32, String)'
        data = [((1, 'a'), ), ((2, 'b'), )]

        with self.create_table(columns):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, "(1,'a')\n(2,'b')\n")

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_tuple_single_element(self):
        columns = 'a Tuple(Int32)'
        data = [((1, ), ), ((2, ), )]

        with self.create_table(columns):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, "(1)\n(2)\n")

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_nullable(self):
        data = [
            ((1, 'a'),),
            ((2, None),), ((None, None),), ((None, 'd'),),
            ((5, 'e'),)
        ]

        with self.create_table('a Tuple(Nullable(Int32), Nullable(String))'):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                "(1,'a')\n"
                "(2,NULL)\n(NULL,NULL)\n(NULL,'d')\n"
                "(5,'e')\n"
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_nested_tuple_with_common_types(self):
        columns = 'a Tuple(String, Tuple(Int32, String), String)'
        data = [
            (('one', (1, 'a'), 'two'),),
            (('three', (2, 'b'), 'four'),)
        ]

        with self.create_table(columns):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                "('one',(1,'a'),'two')\n"
                "('three',(2,'b'),'four')\n"
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_tuple_of_tuples(self):
        columns = (
            "a Tuple("
            "Tuple(Int32, String),"
            "Tuple(Enum8('hello' = 1, 'world' = 2), Date)"
            ")"
        )
        data = [
            (((1, 'a'), (1, date(2020, 3, 11))),),
            (((2, 'b'), (2, date(2020, 3, 12))),)
        ]

        with self.create_table(columns):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                "((1,'a'),('hello','2020-03-11'))\n"
                "((2,'b'),('world','2020-03-12'))\n"
            )

            inserted = self.client.execute(query)
            self.assertEqual(
                inserted, [
                    (((1, 'a'), ('hello', date(2020, 3, 11))), ),
                    (((2, 'b'), ('world', date(2020, 3, 12))), )
                ]
            )

    def test_tuple_of_arrays(self):
        data = [(([1, 2, 3],),), (([4, 5, 6],),)]

        with self.create_table('a Tuple(Array(Int32))'):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                "([1,2,3])\n([4,5,6])\n"
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    # Bug in Array of Tuple handing before 19.16.13:
    # DESCRIBE TABLE test
    #
    # | a.1  | Array(UInt8) |
    # | a.2  | Array(UInt8) |
    # | a.3  | Array(UInt8) |
    # https://github.com/ClickHouse/ClickHouse/pull/8866
    @require_server_version(19, 16, 13)
    def test_array_of_tuples(self):
        data = [
            ([(1, 2, 3), (4, 5, 6)],),
            ([(7, 8, 9)],),
        ]

        with self.create_table('a Array(Tuple(UInt8, UInt8, UInt8))'):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                "[(1,2,3),(4,5,6)]\n"
                "[(7,8,9)]\n"
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_type_mismatch_error(self):
        columns = 'a Tuple(Int32)'
        data = [('test', )]

        with self.create_table(columns):
            with self.assertRaises(errors.TypeMismatchError):
                self.client.execute('INSERT INTO test (a) VALUES', data)



from tests.testcase import BaseTestCase
from clickhouse_driver import errors

ErrorCodes = errors.ErrorCodes


class NullableTestCase(BaseTestCase):
    def test_simple(self):
        columns = 'a Nullable(Int32)'

        data = [(3, ), (None, ), (2, )]
        with self.create_table(columns):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted, '3\n\\N\n2\n'
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_nullable_inside_nullable(self):
        columns = 'a Nullable(Nullable(Int32))'

        with self.assertRaises(errors.ServerException) as e:
            self.client.execute(
                'CREATE TABLE test ({}) ''ENGINE = Memory'.format(columns)
            )

        self.assertEqual(e.exception.code, ErrorCodes.ILLEGAL_TYPE_OF_ARGUMENT)

    def test_nullable_array(self):
        columns = 'a Nullable(Array(Nullable(Array(Nullable(Int32)))))'

        with self.assertRaises(errors.ServerException) as e:
            self.client.execute(
                'CREATE TABLE test ({}) ''ENGINE = Memory'.format(columns)
            )

        self.assertEqual(e.exception.code, ErrorCodes.ILLEGAL_TYPE_OF_ARGUMENT)



from datetime import date, timedelta

from .base import FormatColumn


epoch_start = date(1970, 1, 1)
epoch_end = date(2149, 6, 6)

epoch_start_date32 = date(1925, 1, 1)
epoch_end_date32 = date(2283, 11, 11)


class DateColumn(FormatColumn):
    ch_type = 'Date'
    py_types = (date, )
    format = 'H'

    min_value = epoch_start
    max_value = epoch_end

    date_lut_days = (epoch_end - epoch_start).days + 1
    date_lut = {x: epoch_start + timedelta(x) for x in range(date_lut_days)}
    date_lut_reverse = {value: key for key, value in date_lut.items()}

    def before_write_items(self, items, nulls_map=None):
        null_value = self.null_value

        date_lut_reverse = self.date_lut_reverse
        min_value = self.min_value
        max_value = self.max_value

        for i, item in enumerate(items):
            if nulls_map and nulls_map[i]:
                items[i] = null_value
                continue

            if type(item) != date:
                item = date(item.year, item.month, item.day)

            if min_value <= item <= max_value:
                items[i] = date_lut_reverse[item]
            else:
                items[i] = 0

    def after_read_items(self, items, nulls_map=None):
        date_lut = self.date_lut

        if nulls_map is None:
            return tuple(date_lut[item] for item in items)
        else:
            return tuple(
                (None if is_null else date_lut[items[i]])
                for i, is_null in enumerate(nulls_map)
            )


class Date32Column(DateColumn):
    ch_type = 'Date32'
    format = 'i'

    min_value = epoch_start_date32
    max_value = epoch_end_date32

    date_lut_days = (epoch_end_date32 - epoch_start).days + 1
    date_lut = {
        x: epoch_start + timedelta(x)
        for x in range((epoch_start_date32 - epoch_start).days, date_lut_days)
    }
    date_lut_reverse = {value: key for key, value in date_lut.items()}



# coding=utf-8
from __future__ import unicode_literals

from tests.testcase import BaseTestCase
from clickhouse_driver import errors


class StringTestCase(BaseTestCase):
    def test_unicode(self):
        data = [('яндекс', )]
        with self.create_table('a String'):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, 'яндекс\n')

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_non_utf(self):
        data = [('яндекс'.encode('koi8-r'), )]
        with self.create_table('a String'):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query, encoding='koi8-r')
            self.assertEqual(inserted, 'яндекс\n')

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_null_byte_in_the_middle(self):
        data = [('a\x00b', )]
        with self.create_table('a String'):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, 'a\\0b\n')

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_nullable(self):
        with self.create_table('a Nullable(String)'):
            data = [(None, ), ('test', ), (None, ), ('nullable', )]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '\\N\ntest\n\\N\nnullable\n')

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_buffer_reader(self):
        data = [('a' * 300, )] * 300
        with self.create_table('a String'):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_compressed_client(self):
        with self.created_client(compression=True) as client:
            data = [('a' * 300, )]
            with self.create_table('a String'):
                client.execute(
                    'INSERT INTO test (a) VALUES', data
                )

                query = 'SELECT * FROM test'

                inserted = client.execute(query)
                self.assertEqual(inserted, data)

    def test_custom_encoding(self):
        settings = {'strings_encoding': 'cp1251'}

        data = [(('яндекс'), ), (('test'), )]
        with self.create_table('a String'):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, settings=settings
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query, encoding='cp1251')
            self.assertEqual(inserted, 'яндекс\ntest\n')

            inserted = self.client.execute(query, settings=settings)
            self.assertEqual(inserted, data)
            self.assertIsInstance(inserted[0][0], str)
            self.assertIsInstance(inserted[1][0], str)

    def test_not_supported_types(self):
        datas = [
            [(bytearray(b'asd'), )],
            [(123, )]
        ]
        with self.create_table('a String'):
            for data in datas:
                with self.assertRaises(errors.TypeMismatchError) as e:
                    self.client.execute(
                        'INSERT INTO test (a) VALUES', data,
                        types_check=True
                    )

                self.assertIn('for column "a"', str(e.exception))

                with self.assertRaises(AttributeError):
                    self.client.execute(
                        'INSERT INTO test (a) VALUES', data
                    )


class ByteStringTestCase(BaseTestCase):
    client_kwargs = {'settings': {'strings_as_bytes': True}}

    def test_not_decoded(self):
        data = [
            (bytes('яндекс'.encode('cp1251')), ),
            (bytes('test'.encode('cp1251')), ),
        ]
        with self.create_table('a String'):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query, encoding='cp1251')
            self.assertEqual(inserted, 'яндекс\ntest\n')

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)
            self.assertIsInstance(inserted[0][0], bytes)
            self.assertIsInstance(inserted[1][0], bytes)

    def test_nullable(self):
        with self.create_table('a Nullable(String)'):
            data = [(None, ), (b'test', ), (None, ), (b'nullable', )]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '\\N\ntest\n\\N\nnullable\n')

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_not_supported_types(self):
        datas = [
            [('asd', )],
            [(bytearray(b'asd'), )],
            [(123, )]
        ]
        with self.create_table('a String'):
            for data in datas:
                with self.assertRaises(errors.TypeMismatchError) as e:
                    self.client.execute(
                        'INSERT INTO test (a) VALUES', data,
                        types_check=True
                    )

                self.assertIn('for column "a"', str(e.exception))

                with self.assertRaises(ValueError) as e:
                    self.client.execute(
                        'INSERT INTO test (a) VALUES', data
                    )

                self.assertIn('bytes object expected', str(e.exception))



import numpy as np

from .base import NumpyColumn
from ..util import get_inner_spec, get_inner_columns


class TupleColumn(NumpyColumn):
    def __init__(self, nested_columns, **kwargs):
        self.nested_columns = nested_columns
        super(TupleColumn, self).__init__(**kwargs)

    def write_data(self, items, buf):
        names = items.dtype.names
        for i, (x, name) in enumerate(zip(self.nested_columns, names)):
            x.write_data(items[name], buf)

    def write_items(self, items, buf):
        return self.write_data(items, buf)

    def read_data(self, n_items, buf):
        data = [x.read_data(n_items, buf) for x in self.nested_columns]
        dtype = [('f{}'.format(i), x.dtype) for i, x in enumerate(data)]
        rv = np.empty(n_items, dtype=dtype)
        for i, x in enumerate(data):
            rv['f{}'.format(i)] = x
        return rv

    def read_items(self, n_items, buf):
        return self.read_data(n_items, buf)


def create_tuple_column(spec, column_by_spec_getter, column_options):
    inner_spec = get_inner_spec('Tuple', spec)
    columns = get_inner_columns(inner_spec)

    return TupleColumn([column_by_spec_getter(x) for x in columns],
                       **column_options)



import socket
from unittest import TestCase, mock

from clickhouse_driver.bufferedreader import BufferedSocketReader


class BufferedReaderTestCase(TestCase):
    def test_overflow_signed_int_string_size(self):
        data = b'\xFF\xFE\xFC\xFE\xFE\xFE\xFE\xFE\x29\x80\x40\x00\x00\x01'

        def recv_into(buf):
            size = len(data)
            buf[0:size] = data
            return size

        with mock.patch('socket.socket') as mock_socket:
            mock_socket.return_value.recv_into.side_effect = recv_into
            reader = BufferedSocketReader(socket.socket(), 1024)

            # Trying to allocate huge amount of memory.
            with self.assertRaises(MemoryError):
                reader.read_strings(5, encoding='utf-8')




from .arraycolumn import create_array_column
from .util import get_inner_spec, get_inner_columns, \
    get_inner_columns_with_types


def create_nested_column(spec, column_by_spec_getter, column_options):
    return create_array_column(
        'Array(Tuple({}))'.format(','.join(get_nested_columns(spec))),
        column_by_spec_getter, column_options
    )


def get_nested_columns(spec):
    inner_spec = get_inner_spec('Nested', spec)
    return get_inner_columns(inner_spec)


def get_columns_with_types(spec):
    inner_spec = get_inner_spec('Nested', spec)
    return get_inner_columns_with_types(inner_spec)



from parameterized import parameterized

from clickhouse_driver import errors
from clickhouse_driver.columns.service import get_column_by_spec
from clickhouse_driver.context import Context

from tests.numpy.testcase import NumpyBaseTestCase


class OtherColumnsTestCase(NumpyBaseTestCase):
    def get_column(self, spec):
        ctx = Context()
        ctx.client_settings = {'strings_as_bytes': False, 'use_numpy': True}
        return get_column_by_spec(spec, {'context': ctx})

    @parameterized.expand([
        ("Enum8('hello' = 1, 'world' = 2)", ),
        ('Decimal(8, 4)', ),
        ('Array(String)', ),
        ('Tuple(String)', ),
        ('SimpleAggregateFunction(any, Int32)', ),
        ('Map(String, String)', ),
        ('Array(LowCardinality(String))', )
    ])
    def test_generic_type(self, spec):
        col = self.get_column(spec)
        self.assertIsNotNone(col)

    def test_get_unknown_column(self):
        with self.assertRaises(errors.UnknownTypeError) as e:
            self.get_column('Unicorn')

        self.assertIn('Unicorn', str(e.exception))



from tests.testcase import BaseTestCase
from clickhouse_driver import errors


class BoolTestCase(BaseTestCase):
    required_server_version = (21, 12)

    def test_simple(self):
        columns = ("a Bool")

        data = [(1,), (0,), (True,), (False,), (None,), ("False",), ("",)]
        with self.create_table(columns):
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted, (
                    'true\n'
                    'false\n'
                    'true\n'
                    'false\n'
                    'false\n'
                    'true\n'
                    'false\n'
                )
            )

            inserted = self.client.execute(query)
            self.assertEqual(
                inserted, [
                    (True, ),
                    (False, ),
                    (True, ),
                    (False, ),
                    (False, ),
                    (True, ),
                    (False, ),
                ]
            )

    def test_errors(self):
        columns = "a Bool"
        with self.create_table(columns):
            with self.assertRaises(errors.TypeMismatchError):
                self.client.execute(
                    'INSERT INTO test (a) VALUES', [(1, )],
                    types_check=True
                )

    def test_nullable(self):
        columns = "a Nullable(Bool)"

        data = [(None, ), (True, ), (False, )]
        with self.create_table(columns):
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted, (
                    '\\N\ntrue\nfalse\n'
                )
            )

            inserted = self.client.execute(query)
            self.assertEqual(
                inserted, [
                    (None, ), (True, ), (False, ),
                ]
            )



import os
from datetime import date, datetime
from unittest.mock import patch

from freezegun import freeze_time

from tests.testcase import BaseTestCase


class DateTestCase(BaseTestCase):
    @freeze_time('2017-03-05 03:00:00')
    def test_do_not_use_timezone(self):
        with self.create_table('a Date'):
            data = [(date(1970, 1, 2), )]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '1970-01-02\n')

            with patch.dict(os.environ, {'TZ': 'US/Hawaii'}):
                inserted = self.client.execute(query)
                self.assertEqual(inserted, data)

    def test_insert_datetime_to_date(self):
        with self.create_table('a Date'):
            testTime = datetime(2015, 6, 6, 12, 30, 54)
            self.client.execute(
                'INSERT INTO test (a) VALUES', [(testTime, )]
            )
            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '2015-06-06\n')

    def test_wrong_date_insert(self):
        with self.create_table('a Date'):
            data = [
                (date(5555, 1, 1), ),
                (date(1, 1, 1), ),
                (date(2149, 6, 7), )
            ]
            self.client.execute('INSERT INTO test (a) VALUES', data)
            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            expected = (
                3*'1970-01-01\n' if self.server_version > (20, 7, 2)
                else 3*'0000-00-00\n'
            )
            self.assertEqual(inserted, expected)

    def test_boundaries(self):
        extended_date = self.server_version > (21, 4)

        with self.create_table('a Date'):
            data = [
                (date(1970, 1, 1), ),
                ((date(2149, 6, 6) if extended_date else date(2106, 2, 7)), )
            ]
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            if extended_date:
                expected = '1970-01-01\n2149-06-06\n'
            else:
                if self.server_version > (20, 7, 2):
                    expected = '1970-01-01\n2106-02-07\n'
                else:
                    expected = '0000-00-00\n2106-02-07\n'
            self.assertEqual(inserted, expected)

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)


class Date32TestCase(BaseTestCase):
    required_server_version = (21, 9)

    def test_wrong_date_insert(self):
        with self.create_table('a Date32'):
            data = [
                (date(5555, 1, 1), ),
                (date(1, 1, 1), ),
                (date(2284, 1, 1), )
            ]
            self.client.execute('INSERT INTO test (a) VALUES', data)
            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '1970-01-01\n1970-01-01\n1970-01-01\n')

    def test_boundaries(self):
        with self.create_table('a Date32'):
            data = [(date(1925, 1, 1), ), (date(2283, 11, 11), )]
            self.client.execute('INSERT INTO test (a) VALUES', data)

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '1925-01-01\n2283-11-11\n')

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)



import numpy as np

from .base import NumpyColumn

# normalize_null_value = False due to float('nan')
# With normalization pandas.isnull will threat float('nan') as NULL value.


class NumpyFloat32Column(NumpyColumn):
    dtype = np.dtype(np.float32)
    ch_type = 'Float32'
    normalize_null_value = False

    def _get_nulls_map(self, items):
        return [x is None for x in items]


class NumpyFloat64Column(NumpyColumn):
    dtype = np.dtype(np.float64)
    ch_type = 'Float64'
    normalize_null_value = False

    def _get_nulls_map(self, items):
        return [x is None for x in items]



import os
from contextlib import contextmanager
from functools import wraps
import logging
from io import StringIO
from time import tzset
from unittest.mock import patch

import tzlocal


def skip_by_server_version(testcase, version_required):
    testcase.skipTest(
        'Minimum revision required: {}'.format(
            '.'.join(str(x) for x in version_required)
        )
    )


def require_server_version(*version_required):
    def check(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            self = args[0]
            self.client.connection.connect()

            current = self.client.connection.server_info.version_tuple()

            if version_required <= current:
                return f(*args, **kwargs)
            else:
                skip_by_server_version(self, version_required)

        return wrapper
    return check


class LoggingCapturer(object):
    def __init__(self, logger_name, level):
        self.old_stdout_handlers = []
        self.logger = logging.getLogger(logger_name)
        self.level = level
        super(LoggingCapturer, self).__init__()

    def __enter__(self):
        buffer = StringIO()

        self.new_handler = logging.StreamHandler(buffer)
        self.logger.addHandler(self.new_handler)
        self.old_logger_level = self.logger.level
        self.logger.setLevel(self.level)

        return buffer

    def __exit__(self, *exc_info):
        self.logger.setLevel(self.old_logger_level)
        self.logger.removeHandler(self.new_handler)


capture_logging = LoggingCapturer


def bust_tzlocal_cache():
    try:
        tzlocal.unix._cache_tz = None
        tzlocal.unix._cache_tz_name = None
    except AttributeError:
        pass

    try:
        tzlocal.win32._cache_tz = None
        tzlocal.unix._cache_tz_name = None
    except AttributeError:
        pass


@contextmanager
def patch_env_tz(tz_name):
    bust_tzlocal_cache()

    # Although in many cases, changing the TZ environment variable may
    # affect the output of functions like localtime() without calling
    # tzset(), this behavior should not be relied on.
    # https://docs.python.org/3/library/time.html#time.tzset
    with patch.dict(os.environ, {'TZ': tz_name}):
        tzset()
        yield

    tzset()



from parameterized import parameterized

try:
    import numpy as np
except ImportError:
    np = None

from tests.numpy.testcase import NumpyBaseTestCase


class FloatTestCase(NumpyBaseTestCase):
    n = 10

    def check_result(self, rv, col_type):
        self.assertArraysEqual(rv[0], np.array(range(self.n)))
        self.assertEqual(rv[0].dtype, col_type)

    def get_query(self, ch_type):
        with self.create_table('a {}'.format(ch_type)):
            data = [np.array(range(self.n))]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, columnar=True
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted, '\n'.join(str(x) for x in data[0]) + '\n'
            )
            return self.client.execute(query, columnar=True)

    def test_float32(self):
        rv = self.get_query('Float32')
        self.check_result(rv, np.float32)

    def test_float64(self):
        rv = self.get_query('Float64')
        self.check_result(rv, np.float64)

    def test_fractional_round_trip(self):
        with self.create_table('a Float32'):
            data = [np.array([0.5, 1.5], dtype=np.float32)]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, columnar=True
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '0.5\n1.5\n')

            inserted = self.client.execute(query, columnar=True)
            self.assertArraysEqual(inserted[0], data[0])

    @parameterized.expand(['Float32', 'Float64'])
    def test_nullable(self, float_type):
        with self.create_table('a Nullable({})'.format(float_type)):
            data = [np.array([np.nan, 0.5, None, 1.5], dtype=object)]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, columnar=True
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, 'nan\n0.5\n\\N\n1.5\n')

            inserted = self.client.execute(query, columnar=True)
            self.assertArraysEqual(
                inserted[0].astype(str), data[0].astype(str)
            )
            self.assertEqual(inserted[0].dtype, object)

    def test_nan(self):
        with self.create_table('a Float32'):
            data = [np.array([float('nan'), 0.5], dtype=np.float32)]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, columnar=True
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, 'nan\n0.5\n')

            inserted = self.client.execute(query, columnar=True)
            self.assertArraysEqual(
                inserted[0].astype(str), data[0].astype(str)
            )
            self.assertEqual(inserted[0].dtype, np.float32)



from decimal import Decimal

from clickhouse_driver import errors
from tests.testcase import BaseTestCase
from tests.util import require_server_version


class DecimalTestCase(BaseTestCase):
    required_server_version = (18, 12, 13)
    stable_support_version = (18, 14, 9)
    trailing_zeros_version = (21, 9)

    def client_kwargs(self, version):
        if version < self.stable_support_version:
            return {'settings': {'allow_experimental_decimal_type': True}}

    def cli_client_kwargs(self):
        if self.stable_support_version > self.server_version:
            return {'allow_experimental_decimal_type': 1}

    def test_simple(self):
        with self.create_table('a Decimal(9, 5)'):
            data = [(Decimal('300.42'), ), (300.42, ), (-300, )]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, types_check=True
            )
            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            if self.server_version < self.trailing_zeros_version:
                expected = '300.42000\n300.42000\n-300.00000\n'
            else:
                expected = '300.42\n300.42\n-300\n'
            self.assertEqual(inserted, expected)
            inserted = self.client.execute(query)
            self.assertEqual(inserted, [
                (Decimal('300.42'), ),
                (Decimal('300.42'), ),
                (Decimal('-300'), )
            ])

    @require_server_version(18, 12, 17)
    def test_different_precisions(self):
        columns = 'a Decimal32(2), b Decimal64(2), c Decimal128(2)'

        with self.create_table(columns):
            data = [(
                Decimal('300.42'),
                # 300.42 + (1 << 34)
                Decimal('17179869484.42'),
                # 300.42 + (1 << 100)
                Decimal('1267650600228229401496703205676.42')
            )]
            self.client.execute(
                'INSERT INTO test (a, b, c) VALUES', data
            )

            # Casting to string saves precision.
            query = (
                'SELECT '
                'CAST(a AS String), CAST(b AS String), CAST(c AS String)'
                'FROM test'
            )
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                '300.42\t'
                '17179869484.42\t'
                '1267650600228229401496703205676.42\n'
            )

            inserted = self.client.execute('SELECT * FROM test')
            self.assertEqual(inserted, data)

    @require_server_version(18, 12, 17)
    def test_different_precisions_negative(self):
        columns = 'a Decimal32(2), b Decimal64(2), c Decimal128(2)'

        with self.create_table(columns):
            data = [(
                Decimal('-300.42'),
                # 300.42 + (1 << 34)
                Decimal('-17179869484.42'),
                # 300.42 + (1 << 100)
                Decimal('-1267650600228229401496703205676.42')
            )]
            self.client.execute(
                'INSERT INTO test (a, b, c) VALUES', data
            )

            # Casting to string saves precision.
            query = (
                'SELECT '
                'CAST(a AS String), CAST(b AS String), CAST(c AS String)'
                'FROM test'
            )
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                '-300.42\t'
                '-17179869484.42\t'
                '-1267650600228229401496703205676.42\n'
            )

            inserted = self.client.execute('SELECT * FROM test')
            self.assertEqual(inserted, data)

    @require_server_version(18, 12, 17)
    def test_max_precisions(self):
        columns = 'a Decimal32(0), b Decimal64(0), c Decimal128(0)'

        with self.create_table(columns):
            data = [(
                Decimal(10**9 - 1),
                Decimal(10**18 - 1),
                Decimal(10**38 - 1)
            ), (
                Decimal(-10**9 + 1),
                Decimal(-10**18 + 1),
                Decimal(-10**38 + 1)
            )]
            self.client.execute(
                'INSERT INTO test (a, b, c) VALUES', data
            )

            # Casting to string saves precision.
            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                '999999999\t'
                '999999999999999999\t'
                '99999999999999999999999999999999999999\n'
                '-999999999\t'
                '-999999999999999999\t'
                '-99999999999999999999999999999999999999\n'
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_nullable(self):
        with self.create_table('a Nullable(Decimal32(3))'):
            data = [(300.42, ), (None, )]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            if self.server_version < self.trailing_zeros_version:
                expected = '300.420\n\\N\n'
            else:
                expected = '300.42\n\\N\n'
            self.assertEqual(inserted, expected)

            inserted = self.client.execute(query)
            self.assertEqual(inserted, [(Decimal('300.42'), ), (None, ), ])

    def test_no_scale(self):
        with self.create_table('a Decimal32(0)'):
            data = [(2147483647, )]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '2147483647\n')

            inserted = self.client.execute(query)
            self.assertEqual(inserted, [(Decimal('2147483647'), )])

    def test_type_mismatch(self):
        data = [(2147483649, )]
        with self.create_table('a Decimal32(0)'):
            with self.assertRaises(errors.TypeMismatchError) as e:
                self.client.execute(
                    'INSERT INTO test (a) VALUES', data, types_check=True
                )

            self.assertIn('2147483649 for column "a"', str(e.exception))

            with self.assertRaises(errors.TypeMismatchError) as e:
                self.client.execute(
                    'INSERT INTO test (a) VALUES', data
                )

            self.assertIn('Column a', str(e.exception))

    def test_preserve_precision(self):
        data = [(1.66, ), (1.15, )]

        with self.create_table('a Decimal(18, 2)'):
            self.client.execute('INSERT INTO test (a) VALUES', data)
            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '1.66\n1.15\n')
            inserted = self.client.execute(query)
            self.assertEqual(inserted, [
                (Decimal('1.66'), ),
                (Decimal('1.15'), )
            ])

    def test_precision_one_sign_after_point(self):
        data = [(1.6, ), (1.0, ), (12312.0, ), (999999.6, )]

        with self.create_table('a Decimal(8, 1)'):
            self.client.execute('INSERT INTO test (a) VALUES', data)
            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            if self.server_version < self.trailing_zeros_version:
                expected = '1.6\n1.0\n12312.0\n999999.6\n'
            else:
                expected = '1.6\n1\n12312\n999999.6\n'
            self.assertEqual(inserted, expected)
            inserted = self.client.execute(query)
            self.assertEqual(inserted, [
                (Decimal('1.6'),),
                (Decimal('1.0'),),
                (Decimal('12312.0'),),
                (Decimal('999999.6'),)
            ])

    def test_truncates_scale(self):
        with self.create_table('a Decimal(9, 4)'):
            data = [(3.14159265358,), (2.7182,)]
            expected = [(Decimal('3.1415'),), (Decimal('2.7182'),)]
            self.client.execute(
                'INSERT INTO test (a) VALUES',
                data,
                types_check=True,
            )
            query = 'SELECT * FROM test'
            inserted = self.client.execute(query)
            self.assertEqual(inserted, expected)


class Decimal256TestCase(BaseTestCase):
    required_server_version = (18, 12, 13)

    def cli_client_kwargs(self):
        return {'allow_experimental_bigint_types': 1}

    @require_server_version(20, 9, 2)
    def test_decimal256(self):
        data = [
            (1.66, ),
            (1.15, ),
            # 300.42 + (1 << 200)
            (Decimal('1606938044258990275541962092341162602522202993782792835301676.42'), ),  # noqa: E501
            (Decimal('-1606938044258990275541962092341162602522202993782792835301676.42'), )  # noqa: E501
        ]

        with self.create_table('a Decimal256(2)'):
            self.client.execute('INSERT INTO test (a) VALUES', data)
            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                '1.66\n'
                '1.15\n'
                '1606938044258990275541962092341162602522202993782792835301676.42\n'  # noqa: E501
                '-1606938044258990275541962092341162602522202993782792835301676.42\n'  # noqa: E501
            )
            inserted = self.client.execute(query)
            self.assertEqual(inserted, [
                (Decimal('1.66'), ),
                (Decimal('1.15'), ),
                (Decimal('1606938044258990275541962092341162602522202993782792835301676.42'), ),  # noqa: E501
                (Decimal('-1606938044258990275541962092341162602522202993782792835301676.42'), )  # noqa: E501
            ])



import zstd

from .base import BaseCompressor, BaseDecompressor
from ..protocol import CompressionMethod, CompressionMethodByte


class Compressor(BaseCompressor):
    method = CompressionMethod.ZSTD
    method_byte = CompressionMethodByte.ZSTD

    def compress_data(self, data):
        return zstd.compress(data)


class Decompressor(BaseDecompressor):
    method = CompressionMethod.ZSTD
    method_byte = CompressionMethodByte.ZSTD

    def decompress_data(self, data, uncompressed_size):
        return zstd.decompress(data)



import re
from collections import namedtuple
from functools import lru_cache

from .cursor import Cursor


class DictCursor(Cursor):
    """
    A cursor that generates results as :class:`dict`.

    ``fetch*()`` methods will return dicts instead of tuples.
    """

    def fetchone(self):
        rv = super(DictCursor, self).fetchone()
        if rv is not None:
            rv = dict(zip(self._columns, rv))
        return rv

    def fetchmany(self, size=None):
        rv = super(DictCursor, self).fetchmany(size=size)
        return [dict(zip(self._columns, x)) for x in rv]

    def fetchall(self):
        rv = super(DictCursor, self).fetchall()
        return [dict(zip(self._columns, x)) for x in rv]


class NamedTupleCursor(Cursor):
    """
    A cursor that generates results as named tuples created by
    :func:`~collections.namedtuple`.

    ``fetch*()`` methods will return named tuples instead of regular tuples, so
    their elements can be accessed both as regular numeric items as well as
    attributes.
    """

    # ascii except alnum and underscore
    _re_clean = re.compile(
        '[' + re.escape(' !"#$%&\'()*+,-./:;<=>?@[\\]^`{|}~') + ']')

    @classmethod
    @lru_cache(512)
    def _make_nt(self, key):
        fields = []
        for s in key:
            s = self._re_clean.sub('_', s)
            # Python identifier cannot start with numbers, namedtuple fields
            # cannot start with underscore.
            if s[0] == '_' or '0' <= s[0] <= '9':
                s = 'f' + s
            fields.append(s)

        return namedtuple('Record', fields)

    def fetchone(self):
        rv = super(NamedTupleCursor, self).fetchone()
        if rv is not None:
            nt = self._make_nt(self._columns)
            rv = nt(*rv)
        return rv

    def fetchmany(self, size=None):
        rv = super(NamedTupleCursor, self).fetchmany(size=size)
        nt = self._make_nt(self._columns)
        return [nt(*x) for x in rv]

    def fetchall(self):
        rv = super(NamedTupleCursor, self).fetchall()
        nt = self._make_nt(self._columns)
        return [nt(*x) for x in rv]




# Drop this when minimum supported version will be 3.7.
try:
    import threading
except ImportError:
    import dummy_threading as threading  # noqa: F401

try:
    # since tzlocal 4.0+
    # this will avoid warning for get_localzone().key
    from tzlocal import get_localzone_name

    def get_localzone_name_compat():
        try:
            return get_localzone_name()
        except Exception:
            return None
except ImportError:
    from tzlocal import get_localzone

    def get_localzone_name_compat():
        try:
            return get_localzone().key
        except AttributeError:
            return get_localzone().zone
        except Exception:
            return None



from contextlib import contextmanager

from clickhouse_driver import errors
from tests.testcase import BaseTestCase


class QueryInfoTestCase(BaseTestCase):

    @property
    def sample_query(self):
        return 'SELECT * FROM test GROUP BY foo ORDER BY foo DESC LIMIT 5'

    @contextmanager
    def sample_table(self):
        with self.create_table('foo UInt8'):
            self.client.execute('INSERT INTO test (foo) VALUES',
                                [(i,) for i in range(42)])
            self.client.reset_last_query()
            yield

    def test_default_value(self):
        self.assertIsNone(self.client.last_query)

    def test_store_last_query_after_execute(self):
        with self.sample_table():
            self.client.execute(self.sample_query)

        last_query = self.client.last_query
        self.assertIsNotNone(last_query)
        self.assertIsNotNone(last_query.profile_info)
        self.assertEqual(last_query.profile_info.rows_before_limit, 42)

        self.assertIsNotNone(last_query.progress)
        self.assertEqual(last_query.progress.rows, 42)
        self.assertEqual(last_query.progress.bytes, 42)
        self.assertEqual(last_query.progress.total_rows, 0)

        self.assertGreater(last_query.elapsed, 0)

    def test_last_query_after_execute_iter(self):
        with self.sample_table():
            list(self.client.execute_iter(self.sample_query))

        last_query = self.client.last_query
        self.assertIsNotNone(last_query)
        self.assertIsNotNone(last_query.profile_info)
        self.assertEqual(last_query.profile_info.rows_before_limit, 42)

        self.assertIsNotNone(last_query.progress)
        self.assertEqual(last_query.progress.rows, 42)
        self.assertEqual(last_query.progress.bytes, 42)
        self.assertEqual(last_query.progress.total_rows, 0)

        self.assertEqual(last_query.elapsed, 0)

    def test_last_query_after_execute_with_progress(self):
        with self.sample_table():
            progress = self.client.execute_with_progress(self.sample_query)
            list(progress)
            progress.get_result()

        last_query = self.client.last_query
        self.assertIsNotNone(last_query)
        self.assertIsNotNone(last_query.profile_info)
        self.assertEqual(last_query.profile_info.rows_before_limit, 42)

        self.assertIsNotNone(last_query.progress)
        self.assertEqual(last_query.progress.rows, 42)
        self.assertEqual(last_query.progress.bytes, 42)
        self.assertEqual(last_query.progress.total_rows, 0)

        self.assertEqual(last_query.elapsed, 0)

    def test_last_query_progress_total_rows(self):
        self.client.execute('SELECT max(number) FROM numbers(10)')

        last_query = self.client.last_query
        self.assertIsNotNone(last_query)
        self.assertIsNotNone(last_query.profile_info)
        self.assertEqual(last_query.profile_info.rows_before_limit, 10)

        self.assertIsNotNone(last_query.progress)
        self.assertEqual(last_query.progress.rows, 10)
        self.assertEqual(last_query.progress.bytes, 80)

        total_rows = 10 if self.server_version > (19, 4) else 0
        self.assertEqual(last_query.progress.total_rows, total_rows)

        self.assertGreater(last_query.elapsed, 0)

    def test_last_query_after_execute_insert(self):
        with self.sample_table():
            self.client.execute('INSERT INTO test (foo) VALUES',
                                [(i,) for i in range(42)])

        last_query = self.client.last_query
        self.assertIsNotNone(last_query)
        self.assertIsNotNone(last_query.progress)
        self.assertEqual(last_query.progress.rows, 0)
        self.assertEqual(last_query.progress.bytes, 0)

        self.assertGreater(last_query.elapsed, 0)

    def test_override_after_subsequent_queries(self):
        query = 'SELECT * FROM test WHERE foo < %(i)s ORDER BY foo LIMIT 5'
        with self.sample_table():
            for i in range(1, 10):
                self.client.execute(query, {'i': i})

                profile_info = self.client.last_query.profile_info
                self.assertEqual(profile_info.rows_before_limit, i)

    def test_reset_last_query(self):
        with self.sample_table():
            self.client.execute(self.sample_query)

        self.assertIsNotNone(self.client.last_query)
        self.client.reset_last_query()
        self.assertIsNone(self.client.last_query)

    def test_reset_on_query_error(self):
        with self.assertRaises(errors.ServerException):
            self.client.execute('SELECT answer FROM universe')

        self.assertIsNone(self.client.last_query)

    def test_progress_info_increment(self):
        self.client.execute(
            'SELECT x FROM ('
            'SELECT number AS x FROM numbers(100000000)'
            ') ORDER BY x ASC LIMIT 10'
        )

        last_query = self.client.last_query
        self.assertIsNotNone(last_query)
        self.assertIsNotNone(last_query.progress)
        self.assertGreater(last_query.progress.rows, 100000000)
        self.assertGreater(last_query.progress.bytes, 800000000)

        total_rows = 100000000 if self.server_version > (19, 4) else 0
        self.assertEqual(last_query.progress.total_rows, total_rows)

    def test_progress_info_ddl(self):
        self.client.execute('DROP TABLE IF EXISTS foo')

        last_query = self.client.last_query
        self.assertIsNotNone(last_query)
        self.assertIsNotNone(last_query.progress)
        self.assertEqual(last_query.progress.rows, 0)
        self.assertEqual(last_query.progress.bytes, 0)

        self.assertGreater(last_query.elapsed, 0)



# coding=utf-8
from __future__ import unicode_literals

from tests.testcase import BaseTestCase
from clickhouse_driver import errors


class FixedStringTestCase(BaseTestCase):
    def test_simple(self):
        data = [('a', ), ('bb', ), ('ccc', ), ('dddd', ), ('я', )]
        with self.create_table('a FixedString(4)'):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted, (
                    'a\\0\\0\\0\n'
                    'bb\\0\\0\n'
                    'ccc\\0\n'
                    'dddd\n'
                    'я\\0\\0\n'
                )
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_non_utf(self):
        data = [('яндекс'.encode('koi8-r'), )]
        with self.create_table('a FixedString(6)'):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query, encoding='koi8-r')
            self.assertEqual(inserted, 'яндекс\n')

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_oversized(self):
        columns = 'a FixedString(4)'

        data = [('aaaaa', )]
        with self.create_table(columns):
            with self.assertRaises(errors.TooLargeStringSize):
                self.client.execute(
                    'INSERT INTO test (a) VALUES', data
                )

        data = [('тест', )]
        with self.create_table(columns):
            with self.assertRaises(errors.TooLargeStringSize):
                self.client.execute(
                    'INSERT INTO test (a) VALUES', data
                )

    def test_nullable(self):
        with self.create_table('a Nullable(FixedString(10))'):
            data = [(None, ), ('test', ), (None, ), ('nullable', )]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                '\\N\ntest\\0\\0\\0\\0\\0\\0\n\\N\nnullable\\0\\0\n'
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_null_byte_in_the_middle(self):
        data = [('test\0test', )]
        with self.create_table('a FixedString(9)'):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_empty(self):
        data = [('',)]
        with self.create_table('a FixedString(5)'):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_custom_encoding(self):
        settings = {'strings_encoding': 'cp1251'}

        data = [(('яндекс'), ), (('test'), )]
        with self.create_table('a FixedString(10)'):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data, settings=settings
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query, encoding='cp1251')
            self.assertEqual(
                inserted,
                'яндекс\\0\\0\\0\\0\n'
                'test\\0\\0\\0\\0\\0\\0\n'
            )

            inserted = self.client.execute(query, settings=settings)
            self.assertEqual(inserted, data)
            self.assertIsInstance(inserted[0][0], str)
            self.assertIsInstance(inserted[1][0], str)

    def test_not_supported_types(self):
        datas = [
            [(bytearray(b'asd'), )],
            [(123, )]
        ]
        with self.create_table('a String'):
            for data in datas:
                with self.assertRaises(errors.TypeMismatchError) as e:
                    self.client.execute(
                        'INSERT INTO test (a) VALUES', data,
                        types_check=True
                    )

                self.assertIn('for column "a"', str(e.exception))

                with self.assertRaises(AttributeError):
                    self.client.execute(
                        'INSERT INTO test (a) VALUES', data
                    )


class ByteFixedStringTestCase(BaseTestCase):
    client_kwargs = {'settings': {'strings_as_bytes': True}}

    def test_oversized(self):
        columns = 'a FixedString(4)'

        data = [(bytes('aaaaa'.encode('utf-8')), )]
        with self.create_table(columns):
            with self.assertRaises(errors.TooLargeStringSize):
                self.client.execute(
                    'INSERT INTO test (a) VALUES', data
                )

        data = [(bytes('тест'.encode('utf-8')), )]
        with self.create_table(columns):
            with self.assertRaises(errors.TooLargeStringSize):
                self.client.execute(
                    'INSERT INTO test (a) VALUES', data
                )

    def test_not_decoded(self):
        data = [
            (bytes('яндекс'.encode('cp1251')), ),
            (bytes('test'.encode('cp1251')), ),
        ]
        with self.create_table('a FixedString(8)'):
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query, encoding='cp1251')
            self.assertEqual(
                inserted, 'яндекс\\0\\0\ntest\\0\\0\\0\\0\n'
            )

            inserted = self.client.execute(query)
            # Assert items with trailing zeros
            self.assertEqual(
                inserted, [
                    ('яндекс'.encode('cp1251') + b'\x00' * 2, ),
                    ('test'.encode('cp1251') + b'\x00' * 4, )
                ]
            )
            self.assertIsInstance(inserted[0][0], bytes)
            self.assertIsInstance(inserted[1][0], bytes)

    def test_nullable(self):
        with self.create_table('a Nullable(FixedString(10))'):
            data = [
                (None, ),
                (b'test\x00\x00\x00\x00\x00\x00', ),
                (None, ),
                (b'nullable\x00\x00', )
            ]
            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(
                inserted,
                '\\N\ntest\\0\\0\\0\\0\\0\\0\n\\N\nnullable\\0\\0\n'
            )

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test_not_supported_types(self):
        datas = [
            [('asd', )],
            [(bytearray(b'asd'), )],
            [(123, )]
        ]
        with self.create_table('a String'):
            for data in datas:
                with self.assertRaises(errors.TypeMismatchError) as e:
                    self.client.execute(
                        'INSERT INTO test (a) VALUES', data,
                        types_check=True
                    )

                self.assertIn('for column "a"', str(e.exception))

                with self.assertRaises(ValueError) as e:
                    self.client.execute(
                        'INSERT INTO test (a) VALUES', data
                    )

                self.assertIn('bytes object expected', str(e.exception))




class OpenTelemetryTraceContext(object):
    traceparent_tpl = 'xx-xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx-xxxxxxxxxxxxxxxx-xx'
    translation = str.maketrans('1234567890abcdef', 'xxxxxxxxxxxxxxxx')

    def __init__(self, traceparent, tracestate):
        # xx-xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx-xxxxxxxxxxxxxxxx-xx
        # ^              ^                           ^         ^
        # version     trace_id                    span_id      flags

        self.trace_id = None  # UUID
        self.span_id = None  # UInt64
        self.tracestate = tracestate  # String
        self.trace_flags = None  # UInt8

        if traceparent is not None:
            self.parse_traceparent(traceparent)

        super(OpenTelemetryTraceContext, self).__init__()

    def parse_traceparent(self, traceparent):
        traceparent = traceparent.lower()

        if len(traceparent) != len(self.traceparent_tpl):
            raise ValueError('unexpected length {}, expected {}'.format(
                len(traceparent), len(self.traceparent_tpl)
            ))

        if traceparent.translate(self.translation) != self.traceparent_tpl:
            raise ValueError(
                'Malformed traceparant header: {}'.format(traceparent)
            )

        parts = traceparent.split('-')
        version = int(parts[0], 16)
        if version != 0:
            raise ValueError(
                'unexpected version {}, expected 00'.format(parts[0])
            )

        self.trace_id = (int(parts[1][16:], 16) << 64) + int(parts[1][:16], 16)
        self.span_id = int(parts[2], 16)
        self.trace_flags = int(parts[3], 16)



import os
import re
from codecs import open

from setuptools import setup, find_packages
from distutils.extension import Extension

try:
    from Cython.Build import cythonize
except ImportError:
    USE_CYTHON = False
else:
    USE_CYTHON = True

CYTHON_TRACE = bool(os.getenv('CYTHON_TRACE', False))

here = os.path.abspath(os.path.dirname(__file__))


def read_version():
    regexp = re.compile(r'^VERSION\W*=\W*\(([^\(\)]*)\)')
    init_py = os.path.join(here, 'clickhouse_driver', '__init__.py')
    with open(init_py, encoding='utf-8') as f:
        for line in f:
            match = regexp.match(line)
            if match is not None:
                return match.group(1).replace(', ', '.')
        else:
            raise RuntimeError(
                'Cannot find version in clickhouse_driver/__init__.py'
            )


with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# Prepare extensions.
ext = '.pyx' if USE_CYTHON else '.c'
extensions = [
    Extension(
        'clickhouse_driver.bufferedreader',
        ['clickhouse_driver/bufferedreader' + ext]
    ),
    Extension(
        'clickhouse_driver.bufferedwriter',
        ['clickhouse_driver/bufferedwriter' + ext]
    ),
    Extension(
        'clickhouse_driver.columns.largeint',
        ['clickhouse_driver/columns/largeint' + ext]
    ),
    Extension(
        'clickhouse_driver.varint',
        ['clickhouse_driver/varint' + ext]
    )
]

if USE_CYTHON:
    compiler_directives = {'language_level': '3'}
    if CYTHON_TRACE:
        compiler_directives['linetrace'] = True

    extensions = cythonize(extensions, compiler_directives=compiler_directives)

github_url = 'https://github.com/mymarilyn/clickhouse-driver'

setup(
    name='clickhouse-driver',
    version=read_version(),

    description='Python driver with native interface for ClickHouse',
    long_description=long_description,

    url=github_url,

    author='Konstantin Lebedev',
    author_email='kostyan.lebedev@gmail.com',

    license='MIT',

    classifiers=[
        'Development Status :: 4 - Beta',


        'Environment :: Console',


        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',


        'License :: OSI Approved :: MIT License',


        'Operating System :: OS Independent',


        'Programming Language :: SQL',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: PyPy',

        'Topic :: Database',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Information Analysis'
    ],

    keywords='ClickHouse db database cloud analytics',

    project_urls={
        'Documentation': 'https://clickhouse-driver.readthedocs.io',
        'Changes': github_url + '/blob/master/CHANGELOG.md'
    },
    packages=find_packages('.', exclude=['tests*']),
    python_requires='>=3.6, <4',
    install_requires=[
        'pytz',
        'tzlocal'
    ],
    ext_modules=extensions,
    extras_require={
        'lz4': [
            'lz4<=3.0.1; implementation_name=="pypy"',
            'lz4; implementation_name!="pypy"',
            'clickhouse-cityhash>=1.0.2.1'
        ],
        'zstd': ['zstd', 'clickhouse-cityhash>=1.0.2.1'],
        'numpy': ['numpy>=1.12.0', 'pandas>=0.24.0']
    },
    test_suite='pytest'
)



from datetime import date, datetime
from unittest import TestCase

from clickhouse_driver import errors
from clickhouse_driver.client import Client
from clickhouse_driver.compression import get_compressor_cls
from clickhouse_driver.compression.lz4 import Compressor
from .testcase import BaseTestCase, file_config


class BaseCompressionTestCase(BaseTestCase):
    compression = False
    supported_compressions = file_config.get('db', 'compression').split(',')

    def _create_client(self):
        settings = None
        if self.compression:
            # Set server compression method explicitly
            # By default server sends blocks compressed by LZ4.
            method = self.compression
            if self.server_version > (19, ):
                method = method.upper()
            settings = {'network_compression_method': method}

        return Client(
            self.host, self.port, self.database, self.user, self.password,
            compression=self.compression, settings=settings
        )

    def setUp(self):
        super(BaseCompressionTestCase, self).setUp()
        supported = (
            self.compression is False or
            self.compression in self.supported_compressions
        )

        if not supported:
            self.skipTest(
                'Compression {} is not supported'.format(self.compression)
            )

    def run_simple(self):
        with self.create_table('a Date, b DateTime'):
            data = [(date(2012, 10, 25), datetime(2012, 10, 25, 14, 7, 19))]
            self.client.execute(
                'INSERT INTO test (a, b) VALUES', data
            )

            query = 'SELECT * FROM test'
            inserted = self.emit_cli(query)
            self.assertEqual(inserted, '2012-10-25\t2012-10-25 14:07:19\n')

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)

    def test(self):
        if self.compression is False:
            return

        self.run_simple()


class LZ4ReadWriteTestCase(BaseCompressionTestCase):
    compression = 'lz4'


class LZ4HCReadWriteTestCase(BaseCompressionTestCase):
    compression = 'lz4hc'


class ZSTDReadWriteTestCase(BaseCompressionTestCase):
    compression = 'zstd'


class MiscCompressionTestCase(TestCase):
    def test_default_compression(self):
        client = Client('localhost', compression=True)
        self.assertEqual(client.connection.compressor_cls, Compressor)

    def test_unknown_compressor(self):
        with self.assertRaises(errors.UnknownCompressionMethod) as e:
            get_compressor_cls('hello')

        self.assertEqual(
            e.exception.code, errors.ErrorCodes.UNKNOWN_COMPRESSION_METHOD
        )


class ReadByBlocksTestCase(BaseCompressionTestCase):
    compression = 'lz4'

    def test(self):
        with self.create_table('a Int32'):
            data = [(x % 200, ) for x in range(1000000)]

            self.client.execute(
                'INSERT INTO test (a) VALUES', data
            )

            query = 'SELECT * FROM test'

            inserted = self.client.execute(query)
            self.assertEqual(inserted, data)



from functools import wraps
from unittest import SkipTest


def check_numpy(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except RuntimeError as e:
            if 'NumPy' in str(e):
                raise SkipTest('Numpy package is not installed')

    return wrapper
